#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains save widget for data items implementation for Maya
"""

from __future__ import print_function, division, absolute_import

from Qt.QtGui import QCursor

from tpDcc.managers import resources
from tpDcc.libs.qt.widgets import buttons

from tpDcc.tools.datalibrary.widgets import save
from tpDcc.tools.datalibrary.dccs.maya.menus import sets


class MayaSaveWidget(save.BaseSaveWidget):
    def __init__(self, item_view, *args, **kwargs):
        super(MayaSaveWidget, self).__init__(item_view=item_view, *args, **kwargs)

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def ui(self):
        super(MayaSaveWidget, self).ui()

        self._selection_set_button = buttons.BaseButton(parent=self)
        self._selection_set_button.setIcon(resources.icon('group_objects'))
        self._preview_buttons_layout.insertWidget(2, self._selection_set_button)

    def setup_signals(self):
        super(MayaSaveWidget, self).setup_signals()

        self._selection_set_button.clicked.connect(self._on_show_selection_sets_menu)

    # ============================================================================================================
    # CALLBACKS
    # ============================================================================================================

    def _on_show_selection_sets_menu(self):
        """
        Internal callback function that shows the selection sets menu for the current folder path
        """

        path = self.folder_path()
        position = QCursor.pos()
        library_window = self.library_window()

        menu = sets.SetsMenu.from_path(path, library_window=library_window)
        menu.exec_(position)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains base preview widget for data items
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import Qt, QSize
from Qt.QtWidgets import QWidget, QFrame, QMenu
from Qt.QtGui import QCursor

from tpDcc.managers import resources
from tpDcc.libs.qt.core import base
from tpDcc.libs.qt.widgets import layouts, label, buttons, formwidget, group

from tpDcc.tools.datalibrary.widgets import sequence


class PreviewWidget(base.BaseWidget, object):
    def __init__(self, item_view, parent=None):

        self._item_view = item_view
        self._sequence_widget = None
        self._form_widget = None

        super(PreviewWidget, self).__init__(parent=parent)

        self._create_sequence_widget()

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def get_main_layout(self):
        main_layout = layouts.VerticalLayout(spacing=2, margins=(0, 0, 0, 0))
        main_layout.setAlignment(Qt.AlignTop)

        return main_layout

    def ui(self):
        super(PreviewWidget, self).ui()

        title_frame = QFrame(self)
        title_frame_layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
        title_frame.setLayout(title_frame_layout)
        title_widget = QWidget(self)
        title_layout = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))
        title_widget.setLayout(title_layout)
        buttons_layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
        title_layout.addLayout(buttons_layout)
        self._title_icon = label.BaseLabel(parent=self)
        self._title_button = label.ElidedLabel(parent=self)
        self._title_button.setText(self.item().name())
        self._menu_button = buttons.BaseButton(parent=self)
        self._menu_button.setIcon(resources.icon('menu_dots'))
        buttons_layout.addWidget(self._title_icon)
        buttons_layout.addStretch()
        buttons_layout.addWidget(self._title_button)
        buttons_layout.addStretch()
        buttons_layout.addWidget(self._menu_button)
        title_frame_layout.addWidget(title_widget)

        item_icon_name = self.item().icon() or 'tpDcc'
        item_icon = resources.icon(item_icon_name)
        if not item_icon:
            item_icon = resources.icon('tpDcc')
        self._title_icon.setPixmap(item_icon.pixmap(QSize(20, 20)))

        icon_title_frame = QFrame(self)
        icon_title_frame_layout = layouts.VerticalLayout(spacing=0, margins=(4, 2, 4, 2))
        icon_title_frame.setLayout(icon_title_frame_layout)

        self._icon_group_frame = QFrame(self)
        icon_group_layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
        self._icon_group_frame.setLayout(icon_group_layout)
        self._icon_frame = QFrame(self)
        icon_frame_layout = layouts.VerticalLayout(spacing=0, margins=(4, 2, 4, 2))
        self._icon_frame.setLayout(icon_frame_layout)
        icon_group_layout.addWidget(self._icon_frame)
        icon_group_box_widget = group.GroupBoxWidget('Icon', widget=self._icon_group_frame, parent=self)
        icon_title_frame_layout.addWidget(icon_group_box_widget)

        form_frame = QFrame(self)
        form_frame_layout = layouts.VerticalLayout(spacing=0, margins=(4, 2, 4, 2))
        form_frame.setLayout(form_frame_layout)

        buttons_frame = QFrame(self)
        buttons_frame_layout = layouts.HorizontalLayout(spacing=4, margins=(4, 4, 4, 4))
        buttons_frame.setLayout(buttons_frame_layout)

        self._accept_button = buttons.BaseButton('Load')
        self._accept_button.setIcon(resources.icon('open'))
        self._accept_button.setVisible(False)
        buttons_frame_layout.addWidget(self._accept_button)

        schema = self._item_view.item.load_schema() if self._item_view else None
        if schema:
            self._form_widget = formwidget.FormWidget(self)
            self._form_widget.setObjectName('{}Form'.format(self._item_view.__class__.__name__))
            self._form_widget.set_schema(schema)
            self._form_widget.set_validator(self.validator)
            form_frame_layout.addWidget(self._form_widget)

        self.main_layout.addWidget(title_frame)
        self.main_layout.addWidget(icon_title_frame)
        self.main_layout.addWidget(self._icon_group_frame)
        self.main_layout.addWidget(form_frame)
        self.main_layout.addStretch()
        self.main_layout.addWidget(buttons_frame)

    def setup_signals(self):
        self._menu_button.clicked.connect(self._on_show_menu)
        self._accept_button.clicked.connect(self.on_load)

    def resizeEvent(self, event):
        """
        Overrides base BaseWidget resizeEvent function to make sure the icon image size is updated
        :param event: QSizeEvent
        """

        self.update_thumbnail_size()

    def close(self):
        """
        Overrides base BaseWidget close function to save widget persistent data
        """

        # TODO: It is not sure this function it is going to be called when the app is closed
        # TODO: Find a better approach

        if self._form_widget:
            self._form_widget.save_persistent_values()

        super(PreviewWidget, self).close()

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def item(self):
        """
        Returns the library item to load
        :return: LibraryItem
        """

        return self._item_view

    def validator(self, **kwargs):
        """
        Returns validator used for validating the item loaded arguments
        :param kwargs: dict
        """

        self._item_view.load_validator(**kwargs)
        self.update_icon()

    def update_icon(self):
        """
        Updates item thumbnail icon
        """

        # Updates static icon
        icon = self._item_view.thumbnail_icon()
        if icon:
            self._sequence_widget.setIcon(icon)

        # Update images sequence (if exist)
        if self._item_view.image_sequence_path():
            self._sequence_widget.set_dirname(self._item_view.image_sequence_path())

    def update_thumbnail_size(self):
        """
        Updates the thumbnail button to the size of the widget
        """

        width = self.width() - 5
        width = width if width <= 150 else 150
        size = QSize(width, width)
        self._icon_frame.setMaximumSize(size)
        self._icon_group_frame.setMaximumSize(size)

        if self._sequence_widget:
            self._sequence_widget.setIconSize(size)
            self._sequence_widget.setMinimumSize(size)
            self._sequence_widget.setMaximumSize(size)

    # ============================================================================================================
    # INTERNAL
    # ============================================================================================================

    def _create_sequence_widget(self):
        """
        Internal function that creates sequence widget for the item
        """

        self._sequence_widget = sequence.ImageSequenceWidget(self._icon_frame)
        self._icon_frame.layout().insertWidget(0, self._sequence_widget)
        self.update_icon()

    # ============================================================================================================
    # CALLBACKS
    # ============================================================================================================

    def _on_item_data_changed(self, *args, **kwargs):
        """
        Internal callback function that is called when item data changes
        :param args:
        :param kwargs:
        """

        self.update_icon()

    def _on_show_menu(self):
        """
        Internal callback function triggered when item menu should be opened
        :return: QAction
        """

        menu = QMenu(self)
        self.item().context_edit_menu(menu)
        point = QCursor.pos()
        point.setX(point.x() + 3)
        point.setY(point.y() + 3)

        return menu.exec_(point)

    def on_load(self):
        """
        Internal callback function that is triggered when load button is pressed by the user
        """

        kwargs = self._form_widget.values()
        self._item_view.load(**kwargs)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains library data item widget implementation
"""

from __future__ import print_function, division, absolute_import

import os

from tpDcc.libs.datalibrary.data import png, jpg

from tpDcc.tools.datalibrary.data import base


class ImageItemView(base.BaseDataItemView):

    REPRESENTING = [png.PngImageData.__name__, jpg.JpgImageData.__name__]

    NAME = 'Image View'

    def thumbnail_path(self):

        item_path = self.item.format_identifier()
        if os.path.isfile(item_path):
            return item_path

        return self._default_thumbnail_path



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains library item widget implementation
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import Qt, QRect, QSize
from Qt.QtWidgets import QTreeWidgetItem
from Qt.QtGui import QFontMetrics, QColor, QPen, QBrush

from tpDcc.tools.datalibrary.core import consts
from tpDcc.tools.datalibrary.core.views import item


class GroupDataItemView(item.ItemView, object):
    """
    Class that defines group of items
    """

    NAME = 'Group View'

    DEFAULT_FONT_SIZE = consts.GROUP_ITEM_DEFAULT_FONT_SIZE
    PADDING_LEFT = consts.GROUP_ITEM_PADDING_LEFT
    PADDING_RIGHT = consts.GROUP_ITEM_PADDING_RIGHT
    HEIGHT = consts.GROUP_ITEM_HEIGHT

    def __init__(self, *args):
        super(GroupDataItemView, self).__init__(data_item=None, *args)

        self._children = list()

        self._font = self.font(0)
        self._font.setBold(True)

        self._data = dict()

        self.setFont(0, self._font)
        self.setFont(1, self._font)
        self.set_drag_enabled(False)

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def textAlignment(self, column):
        """
        Overrides base LibraryItem textAlignment function
        :param column: int
        """

        return QTreeWidgetItem.textAlignment(self, column)

    def sizeHint(self, column=0):
        """
        Overrides base sizeHint textAlignment function
        Returns the size hint of the item
        :param column: int
        :return: QSize
        """

        padding = self.PADDING_RIGHT * self.dpi()
        width = self.viewer().width() - padding
        return QSize(width, self.HEIGHT * self.dpi())

    def visualRect(self, option):
        """
        Overrides base sizeHint visualRect function
        Returns the visual rect for the item
        :param option: QStyleOptionViewItem
        :return: QRect
        """
        rect = QRect(option.rect)
        rect.setX(self.PADDING_LEFT * self.dpi())
        rect.setWidth(self.sizeHint().width())
        return rect

    def backgroundColor(self):
        """
        Overrides base sizeHint backgroundColor function
        Return the background color for the item.
        :rtype: QtWidgets.QtColor
        """
        return QColor(0, 0, 0, 0)

    def icon(*args):
        """
        Overrides base sizeHint icon function
        Override so icon is not displayed
        :param args:
        :return:
        """
        return None

    def is_label_over_item(self):
        """
        Override function to ignore this feature for group items
        :return: bool
        """

        return False

    def is_label_under_item(self):
        """
        Override function to ignore this feature for group items
        :return: bool
        """

        return False

    def paint_row(self, painter, option, index):
        """
        Overrides base paint_row icon function
        Paint performs low-level painting for the item
        :param painter: QPainter
        :param option: QStyleOptionViewItem
        :param index: QModelIndex
        """

        self.set_rect(QRect(option.rect()))
        painter.save()
        try:
            self.paint_background(painter, option, index)
            if self.is_text_visible():
                self._paint_text(painter, option, 1)
            # self.paint_icon(painter, option, index)
        finally:
            painter.restore()

    def paint_background(self, painter, option, index):
        """
        Overrides base paint_background icon function
        Draw the background for the item
        :param painter: QPainter
        :param option: QStyleOptionViewItem
        :param index: QModelIndex
        """

        super(GroupDataItemView, self).paint_background(painter, option, index)

        painter.setPen(QPen(Qt.NoPen))
        visual_rect = self.visualRect(option)
        text = self.name()
        metrics = QFontMetrics(self._font)
        text_width = metrics.width(text)
        padding = (25 * self.dpi())
        visual_rect.setX(text_width + padding)
        visual_rect.setY(visual_rect.y() + (visual_rect.height() / 2))
        visual_rect.setHeight(2 * self.dpi())
        visual_rect.setWidth(visual_rect.width() - padding)

        color = QColor(self.text_color().red(), self.text_color().green(), self.text_color().blue(), 10)
        painter.setBrush(QBrush(color))
        painter.drawRect(visual_rect)

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def name(self):
        return self._data.get('name')

    def set_name(self, name):
        self._data['name'] = name

    def children(self):
        """
        Returns the children for the group
        :return: list(LibraryItem)
        """

        return self._children

    def set_children(self, children):
        """
        Sets the children for the group
        :param children: list(LibraryItem)
        """

        self._children = children

    def children_hidden(self):
        """
        Returns whether children are hidden or not
        :return: bool
        """

        for child in self.children():
            if not child.isHidden():
                return False

        return True

    def update_children(self):
        """
        Updates the visibility if all children are hidden
        :return:
        """

        self.setHidden(bool(self.children_hidden()))

    def is_text_visible(self):
        """
        Returns whether the text is visible or not
        :return: bool
        """

        return True

    def text_selected_color(self):
        """
        Returns the selected text color for the item
        :return: QColor
        """

        return self.viewer().text_color()

    def background_hover_color(self):
        """
        Return the background color when the mouse is over the item.
        :rtype: QtWidgets.QtColor
        """
        return QColor(0, 0, 0, 0)

    def background_selected_color(self):
        """
        Return the background color when the item is selected.
        :rtype: QtWidgets.QtColor
        """
        return QColor(0, 0, 0, 0)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains data library group by menu widget implementation
"""

from __future__ import print_function, division, absolute_import

from functools import partial

from Qt.QtWidgets import QMenu
from Qt.QtGui import QCursor

from tpDcc.libs.qt.widgets import action


class GroupByMenu(QMenu, object):
    def __init__(self, *args, **kwargs):
        super(GroupByMenu, self).__init__(*args, **kwargs)

        self._library = None

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def show(self, point=None):
        """
        Overrides base QMenu show function
        :param point: QPoint
        """

        self.clear()

        group_by = self.library().group_by()
        if group_by:
            current_field = group_by[0].split(':')[0]
            current_order = 'dsc' if 'dsc' in group_by[0]else 'asc'
        else:
            current_field = ''
            current_order = ''

        separator_action = action.SeparatorAction('Group By', self)
        self.addAction(separator_action)

        none_action = self.addAction('None')
        none_action.setCheckable(True)
        none_callback = partial(self.set_group_by, None, None)
        none_action.triggered.connect(none_callback)
        if not current_field:
            none_action.setChecked(True)

        fields = self.library().fields()
        for field in fields:
            if not field.get('groupable'):
                continue
            name = field.get('name')
            field_action = self.addAction(name.title())
            field_action.setCheckable(True)
            field_action.setChecked(bool(current_field == name))
            field_callback = partial(self.set_group_by, name, current_order)
            field_action.triggered.connect(field_callback)

        group_order_action = action.SeparatorAction('Group Order', self)
        self.addAction(group_order_action)

        ascending_action = self.addAction('Ascending')
        ascending_action.setCheckable(True)
        ascending_action.setChecked(bool(current_order == 'asc'))
        ascending_callback = partial(self.set_group_by, current_field, 'asc')
        ascending_action.triggered.connect(ascending_callback)

        descending_action = self.addAction('Descending')
        descending_action.setCheckable(True)
        descending_action.setChecked(bool(current_order == 'dsc'))
        descending_callback = partial(self.set_group_by, current_field, 'dsc')
        descending_action.triggered.connect(descending_callback)

        point = point or QCursor.pos()

        self.exec_(point)

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def library(self):
        """
        Returns the library model for the menu
        :return: Library
        """

        return self._library

    def set_library(self, library):
        """
        Set the library model for the menu
        :param library: Library
        """

        self._library = library

    def set_group_by(self, group_name, group_order):
        """
        Sets the group by value for the library
        :param group_name: str
        :param group_order: str
        """

        if group_name:
            value = [group_name + ':' + group_order]
        else:
            value = None

        self.library().set_group_by(value)
        self.library().search()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that defines toolset widget implementation for tpDcc-tools-datalibrary
"""

from __future__ import print_function, division, absolute_import

import os

from tpDcc.libs.qt.widgets import toolset
from tpDcc.tools.datalibrary.widgets import window
from tpDcc.tools.datalibrary.widgets import settings


class DataLibraryToolset(toolset.ToolsetWidget, object):
    def __init__(self, *args, **kwargs):
        super(DataLibraryToolset, self).__init__(*args, **kwargs)

        self._repository_widget = settings.DataRepositoryWidget(parent=self._preferences_widget)
        self._preferences_widget.add_category(self._repository_widget.CATEGORY, self._repository_widget)

    @property
    def library_window(self):
        return self._data_library_window

    def post_content_setup(self):
        if not self.client:
            return

        item_paths = self.client.load_data_items() or list()

        paths_to_register = list()
        for item_path in item_paths:
            if not item_path or not os.path.isdir(item_path):
                continue
            paths_to_register.append(item_path)

        library = self._data_library_window.library()
        if not library:
            return

        for path_to_register in paths_to_register:
            library.register_plugin_path(path_to_register)

    def contents(self):
        self._data_library_window = window.LibraryWindow(settings=self.settings, parent=self, client=self._client)
        return [self._data_library_window]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains sidebar data library widget implementation
"""

from __future__ import print_function, division, absolute_import

import os
import logging
from functools import partial
from collections import OrderedDict

from Qt.QtCore import Qt, Signal, QSize, QUrl, QEvent
from Qt.QtWidgets import QFrame, QTreeWidget, QTreeWidgetItem, QAbstractItemView
from Qt.QtGui import QCursor, QColor, QPixmap, QPainter, QBrush

from tpDcc.managers import resources
from tpDcc.libs.python import python, path as path_utils
from tpDcc.libs.resources.core import color, pixmap
from tpDcc.libs.qt.core import base, menu, contexts as qt_contexts
from tpDcc.libs.qt.widgets import layouts, buttons, search

LOGGER = logging.getLogger('tpDcc-libs-datalibrary')


class SidebarWidget(base.BaseWidget):

    itemDropped = Signal(object)
    itemRenamed = Signal(str, str)
    settingsMenuRequested = Signal(object)

    def __init__(self, parent=None):

        self._library = None
        self._previous_filter_text = ''

        super(SidebarWidget, self).__init__(parent=parent)

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def get_main_layout(self):
        return layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))

    def ui(self):
        super(SidebarWidget, self).ui()

        self._title_widget = QFrame(self)
        title_layout = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))
        self._title_widget.setLayout(title_layout)

        buttons_layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
        self._title_button = buttons.BaseButton(parent=self)
        self._title_button.setIcon(resources.icon('reset'))
        self._menu_button = buttons.BaseButton(parent=self)
        self._menu_button.setIcon(resources.icon('menu_dots'))
        buttons_layout.addWidget(self._title_button)
        buttons_layout.addStretch()
        buttons_layout.addWidget(self._menu_button)

        self._filter_search = search.SearchFindWidget(parent=self)
        self._filter_search.setVisible(False)

        title_layout.addLayout(buttons_layout)
        title_layout.addWidget(self._filter_search)

        self._tree_widget = SidebarTree(self)
        self._tree_widget.installEventFilter(self)
        self._tree_widget.itemDropped = self.itemDropped
        self._tree_widget.itemRenamed = self.itemRenamed
        self.itemSelectionChanged = self._tree_widget.itemSelectionChanged

        self._filter_search.set_text(self._tree_widget.filter_text())

        self.main_layout.addWidget(self._title_widget)
        self.main_layout.addWidget(self._tree_widget)

    def setup_signals(self):
        self._title_button.clicked.connect(self.clear_selection)
        self._menu_button.clicked.connect(self._on_show_settings_menu)
        self._filter_search.textChanged.connect(self._on_search_changed)

    def eventFilter(self, obj, event):
        if event.type() == QEvent.KeyPress:
            text = event.text().strip()
            if text.isalpha() or text.isdigit():
                if text and not self._filter_search.search_line.hasFocus():
                    self._filter_search.set_text(text)
                self.set_filter_visible(True)
                self._previous_filter_text = text

        return super(SidebarWidget, self).eventFilter(obj, event)

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def library(self):
        """
        Returns the library model for the menu
        :return: Library
        """

        return self._library

    def set_library(self, library):
        """
        Set the library model for the menu
        :param library: Library
        """

        if library == self._library:
            return

        self._library = library
        # self._library.dataChanged.connect(self._on_data_changed)
        # self._on_data_changed()

    def set_filter_visible(self, flag):
        """
        Sets whether or not filter widget is visible
        :param flag: bool
        """

        self._filter_search.setVisible(flag)
        self._filter_search.search_line.setFocus()
        if not flag and bool(self._tree_widget.filter_text()):
            self._tree_widget.set_filter_text('')
        else:
            self.refresh_filter()

    def search(self):
        if not self.library():
            LOGGER.info('No library found for sidebar widget')
            return

        self.library().add_query(self.query())
        self.library().search()

    def query(self):
        """
        Returns the query for the sidebar widget
        :return: dict
        """

        filters = list()

        for path in self.selected_paths():
            if self.is_recursive():
                filter_ = ('path', 'startswith', path + '/')
                filters.append(filter_)
            filter_ = ('path', 'not', path)
            filters.append(filter_)
        unique_name = 'sidebar_widget_' + str(id(self))
        return {'name': unique_name, 'operator': 'and', 'filters': filters}

    # ============================================================================================================
    # TREE HELPERS FUNCTIONS
    # ============================================================================================================

    def tree_widget(self):
        return self._tree_widget

    def set_dpi(self, dpi):
        self._tree_widget.set_dpi(dpi)

    def is_root_visible(self):
        return self._tree_widget.is_root_visible()

    def set_root_visible(self, flag):
        self._tree_widget.set_root_visible(flag)

    def filter_text(self):
        return self._tree_widget.filter_text()

    def set_filter_text(self, text):
        self._filter_search.set_text(text)

    def refresh_filter(self):
        self._tree_widget.set_filter_text(self._filter_search.get_text())

    def is_filter_visible(self):
        return bool(self._tree_widget.filter_text()) or self._filter_search.isVisible()

    def icons_visible(self):
        return self._tree_widget.icons_visible()

    def set_icons_visible(self, flag):
        self._tree_widget.set_icons_visible(flag)

    def is_recursive(self):
        return self._tree_widget.is_recursive()

    def set_recursive(self, flag):
        self._tree_widget.set_recursive(flag)

    def is_locked(self):
        return self._tree_widget.is_locked()

    def set_locked(self, flag):
        self._tree_widget.set_locked(flag)

    def set_data(self, *args, **kwargs):
        self._tree_widget.set_data(*args, **kwargs)

    def set_item_data(self, item_id, item_data):
        self._tree_widget.set_path_settings(item_id, item_data)

    def selected_path(self):
        return self._tree_widget.selected_path()

    def selected_paths(self):
        return self._tree_widget.selected_paths()

    def select_paths(self, paths):
        self._tree_widget.select_paths(paths)

    def clear_selection(self):
        self._tree_widget.clearSelection()

    # ============================================================================================================
    # SETTINGS
    # ============================================================================================================

    def settings(self):
        """
        Returns sidebar widget settings
        :return: dict
        """

        settings = self._tree_widget.settings()
        settings['filterText'] = self.filter_text()
        settings['filterVisible'] = self.is_filter_visible()

        return settings

    def set_settings(self, settings_dict):
        """
        Sets sidebar widget settings
        :param settings_dict: dict
        """

        value = settings_dict.get('filterText')
        if value is not None:
            self.set_filter_text(value)

        self._tree_widget.set_settings(settings_dict)
        value = settings_dict.get('filterVisible')
        if value is not None:
            self.set_filter_visible(value)

    # ============================================================================================================
    # INTERNAL
    # ============================================================================================================

    def _filter_visible_trigger(self, flag):
        """
        Internal function that sets filter visible
        """

        self.set_filter_visible(flag)
        self._filter_search.select_all()

    def _create_settings_menu(self, settings_menu):
        """
        Internal function that creates sidebar settings context menu
        :param settings_menu: Menu
        :return: Menu
        """

        show_filter_action = settings_menu.addAction(resources.icon('filter'), 'Show Filter')
        show_filter_action.setCheckable(True)
        show_filter_action.setChecked(self.is_filter_visible())
        show_filter_action_callback = partial(self._filter_visible_trigger, not self.is_filter_visible())
        show_filter_action.triggered.connect(show_filter_action_callback)

        show_icons_action = settings_menu.addAction(resources.icon('icons'), 'Show Icons')
        show_icons_action.setCheckable(True)
        show_icons_action.setChecked(self.icons_visible())
        show_icons_action_callback = partial(self.set_icons_visible, not self.icons_visible())
        show_icons_action.triggered.connect(show_icons_action_callback)

        show_root_folder_action = settings_menu.addAction(resources.icon('folder_tree'), 'Show Root Folder')
        show_root_folder_action.setCheckable(True)
        show_root_folder_action.setChecked(self.is_root_visible())
        show_root_folder_action_callback = partial(self.set_root_visible, not self.is_root_visible())
        show_root_folder_action.triggered.connect(show_root_folder_action_callback)

        return settings_menu

    # ============================================================================================================
    # CALLBACKS
    # ============================================================================================================

    def _on_data_changed(self):
        """
        Internal callback function that is triggered when the library data changes
        """

        pass

    def _on_search_changed(self, text):
        """
        Internal callback function called when the search filter has changed
        :param text: str
        """

        self.refresh_filter()
        if text:
            self.set_filter_visible(True)
        else:
            self._tree_widget.setFocus()
            self.set_filter_visible(False)

    def _on_show_settings_menu(self):
        settings_menu = menu.Menu(self)
        self.settingsMenuRequested.emit(settings_menu)
        self._create_settings_menu(settings_menu)
        point = QCursor.pos()
        point.setX(point.x() + 3)
        point.setY(point.y() + 3)
        settings_menu.exec_(point)
        settings_menu.close()


class SidebarTree(QTreeWidget):

    DEFAULT_SEPARATOR = '/'

    itemDropped = Signal(object)
    itemRenamed = Signal(str, str)
    itemSelectionChanged = Signal()     # This is overriding an standard signal.Check if this can cause problems.

    def __init__(self, *args, **kwargs):
        super(SidebarTree, self).__init__(*args, **kwargs)

        self._dpi = 1.0
        self._data = list()
        self._items = list()
        self._index = dict()
        self._locked = False
        self._library = None
        self._recursive = True
        self._filter_text = ''
        self._root_visible = False
        self._icons_visible = True

        self._options = {
            'field': 'path',
            'separator': '/',
            'recursive': True,
            'autoRootPath': True,
            'rootText': 'FOLDERS',
            'sortBy': None,
            'queries': [{'filters': [('type', 'is', 'Folder')]}]
        }

        self.set_dpi(1.0)

        self.setAcceptDrops(True)
        self.setHeaderHidden(True)
        self.setFrameShape(QFrame.NoFrame)
        self.setSelectionMode(QTreeWidget.ExtendedSelection)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)

        self.itemExpanded.connect(self.update)
        self.itemCollapsed.connect(self.update)

    # ============================================================================================================
    # CLASS METHODS
    # ============================================================================================================

    @classmethod
    def find_root(cls, paths, separator=None):
        """
        Finds the common path for the given paths
        :param paths: list(str)
        :param separator: str
        :return: str
        """

        path = paths[0] if paths else ''
        result = None
        separator = separator or cls.DEFAULT_SEPARATOR
        tokens = path.split(separator)
        for i, token in enumerate(tokens):
            root = separator.join(tokens[:i + 1])
            match = True
            for path in paths:
                if not path.startswith(root + separator):
                    match = False
                    break
            if not match:
                break
            result = root

        return result

    @classmethod
    def paths_to_dict(cls, paths, root='', separator=None):
        """
        Returns the given paths as a nested dict
        paths = ['/test/a', '/test/b']
        Result = {'test' : {'a':{}}, {'b':{}}}
        :param paths: list(str)
        :param root: str
        :param separator: str or None
        :return: dict
        """

        separator = separator or cls.DEFAULT_SEPARATOR
        results = OrderedDict()
        paths = path_utils.normalize_paths(paths)

        for path in paths:
            p = results
            # This is to add support for grouping by the given root path.
            if root and root in path:
                path = path.replace(root, "")
                p = p.setdefault(root, OrderedDict())
            keys = path.split(separator)[0:]
            for key in keys:
                if key:
                    p = p.setdefault(key, OrderedDict())

        return results

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def items(self):
        """
        Returns a list of all the items in the tree widget
        :return: list(QTreeWidgetItem)
        """

        items = self.findItems('*', Qt.MatchWildcard | Qt.MatchRecursive)
        return items

    def itemAt(self, pos):
        """
        Overrides base QTreeWidget itemAt function
        :param pos: QPoint
        :return:
        """

        index = self.indexAt(pos)
        if not index.isVaild():
            return None

        return self.itemFromIndex(index)

    def selectionChanged(self, *args):
        """
        Overrides base QTreeWidget selectionChanged function
        :param args:
        :return:
        """

        self.parent().search()

    def dragEnterEvent(self, event):
        """
        Overrides base QTreeWidget dragMoveEvent function
        :param event: QEvent
        """

        mime_data = event.mimeData()
        if mime_data.hasUrls():
            event.accept()
        else:
            event.ignore()

        item = self.itemAt(event.pos())
        if item:
            self.select_paths([item.path()])

    def dragEnterEvent(self, event):
        """
        Overrides base QTreeWidget dragMoveEvent function
        :param event: QEvent
        """

        event.accept()

    def dropEvent(self, event):
        """
        Overrides base QTreeWidget dropEvent function
        :param event: QEvent
        """

        if self.is_locked():
            LOGGER.warning('Folder is locked! Cannot accept drop!')
            return

        self.itemDropped.emit(event)

    def clear(self):
        """
        Clears all the items from the tree widget
        """

        self._items = list()
        self._index = dict()
        super(SidebarTree, self).clear()

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def is_root_visible(self):
        """
        Returns whether or not root root item is visible
        :return: flag
        """

        return self._root_visible

    def set_root_visible(self, flag):
        """
        Sets whether or not root item is visible
        :param flag: bool
        """

        self._root_visible = flag
        self.refresh_data()

    def icons_visible(self):
        """
        Returns whether or not tree item icons are visible
        :return: bool
        """

        return self._icons_visible

    def set_icons_visible(self, flag):
        """
        Sets whether or not tree items icons are visible
        :param flag: bool
        """

        self._icons_visible = flag
        self.refresh_data()

    def is_locked(self):
        """
        Returns whether widget items are read only or not
        :return: bool
        """

        return self._locked

    def set_locked(self, locked):
        """
        Sets whether widget items are read only or not
        :param locked: bool
        """

        self._locked = locked

    def is_recursive(self):
        """
        Returns whether the recursive query is enabled or not
        :return: bool
        """

        return self._recursive

    def set_recursive(self, enable):
        """
        Sets whether recursive query is enabled or not
        :param enable: bool
        """

        self._recursive = enable
        self.parent().search()

    def separator(self):
        """
        Returns the separator used in the fields to seaprate level values
        :return: str
        """

        return self._options.get('separator', self.DEFAULT_SEPARATOR)

    def add_paths(self, paths, root='', split=None):
        """
        Set the given items as a flat list
        :param paths: list(str)
        :param root: str or None
        :param split: str or None
        """

        data = self.paths_to_dict(paths, root=root, separator=split)
        self.create_items(data, split=split)
        if isinstance(paths, dict):
            self.set_settings(paths)

    def select_item(self, item):
        """
        Selects given item
        :param item: QTreeWidgetItem
        """

        self.select_paths([item.path()])

    def selected_item(self):
        """
        Returns the current selected item
        :return: QTreeWidgetItem
        """

        path = self.selected_path()
        return self.item_from_path(path)

    def selected_path(self):
        """
        Returns the current selected path
        :return: str
        """
        paths = self.selected_paths()
        if paths:
            return paths[-1]

    def selected_paths(self):
        """
        Returns the paths that are selected
        :return: list(str)
        """

        paths = list()
        items = self.selectedItems()
        for item in items:
            path = item.path()
            paths.append(path)

        return path_utils.normalize_paths(paths)

    def select_path(self, path):
        """
        Selects the given path
        :param path: str
        """

        self.select_paths([path])

    def select_paths(self, paths):
        """
        Select the items with the given paths
        :param paths: list(str)
        """

        paths = path_utils.normalize_paths(paths)
        items = self.items()
        for item in items:
            if path_utils.clean_path(item.path()) in paths:
                item.setSelected(True)
            else:
                item.setSelected(False)

    def select_url(self, url):
        """
        Select the item with the given url
        :param url: str
        """

        items = self.items()
        for item in items:
            if item.url() == url:
                item.setSelected(True)
            else:
                item.setSelected(False)

    def selected_urls(self):
        """
        Returns the urls for the selected items
        :return: list(str)
        """

        urls = list()
        items = self.selectedItems()
        for item in items:
            urls.append(item.url())

        return urls

    def item_from_url(self, url):
        """
        Returns the item for the given URL
        :param url: QUrl
        :return: QTreeWidgetItem
        """

        for item in self.items():
            if url == item.url():
                return item

    def item_from_path(self, path):
        """
        Returns the item for the given path
        :param path: str
        :return: QTreeWidgetItem
        """

        return self._index.get(path)

    def expanded_items(self):
        """
        Returns all the expanded items
        :return: list(QTreeWidgetItem)
        """

        for item in self.items():
            if self.isItemExpanded(item):
                yield item

    def expanded_paths(self):
        """
        Returns all the expanded paths
        :return: list(QUrl)
        """

        for item in self.expanded_items():
            yield item.url()

    def set_expanded_paths(self, paths):
        """
        Sets the given paths as expanded
        :param paths: list(str)
        """

        for item in self.items():
            if item.url() in paths:
                item.setExpanded(True)

    def set_data(self, data, root='', split=None):
        """
        Sets the items to the given items
        :param data: list(str)
        :param root: str
        :param split: str
        """

        self._data = data
        settings = self.settings()
        with qt_contexts.block_signals(self):
            self.clear()
            if not root:
                root = self.find_root(list(data.keys()), self.separator())
            self.add_paths(data, root=root, split=split)
            self.set_settings(settings)

        self.parent().search()

    def refresh_data(self):
        """
        Updates current tree data
        """

        self.set_data(self._data)

    def create_items(self, data, split=None):
        """
        Creates the items from the given data dict
        :param data: dict
        :param split: str or None
        """

        split = split or self.DEFAULT_SEPARATOR
        self._index = dict()
        for key in data:
            root = split.join([key])
            item = None
            if self.is_root_visible():
                text = key.split(split)
                text = text[-1] if text else key
                item = SidebarTreeItem(self)
                item.setText(0, text)
                item.set_path(root)
                item.setExpanded(True)
                self._index[root] = item

            def _recursive(parent, children, split=None, root=''):
                for text, val in sorted(children.items()):
                    parent = parent or self
                    path = split.join([root, text]).replace('//', '/')

                    # TODO: Maybe we should move this into a filter?
                    # We do not show special folders that starts with '.'
                    if text.startswith('.'):
                        continue

                    child = SidebarTreeItem(parent)
                    child.setText(0, str(text))
                    child.set_path(path)
                    self._index[path] = child
                    _recursive(child, val, split=split, root=path)

            _recursive(item, data[key], split=split, root=root)

        self.update()
        self.refresh_filter()

    def filter_text(self):
        """
        Returns current filter text
        :return: str
        """

        return self._filter_text

    def set_filter_text(self, text):
        """
        Sets current filter text
        :param text: str
        """

        self._filter_text = text.strip()
        self.refresh_filter()

    def refresh_filter(self):
        """
        Refreshes current visible items depending the current filter text
        """

        items = self.items()
        for item in items:
            if self._filter_text.lower() in item.text(0).lower():
                item.setHidden(False)
                for parent in item.parents():
                    parent.setHidden(False)
            else:
                item.setHidden(True)

    def root_text(self):
        """
        Returns the root text
        :return: str
        """

        return self._options.get('rootText')

    def field(self):
        """
        Returns the field
        :return: str
        """

        return self._options.get('field', '')

    def sort_by(self):
        """
        Returns the sort by field
        :return: str
        """

        return self._options.get('sortBy', [self.field()])

    def update(self):
        """
        Updates current tree items
        """

        for item in self.items():
            item.update()

    # ============================================================================================================
    # DPI
    # ============================================================================================================

    def dpi(self):
        """
          Returns the zoom multiplier
          :return: float
          """

        return self._dpi

    def set_dpi(self, dpi):
        """
        Sets zoom multiplier
        :param dpi: float
        """

        self._dpi = dpi
        width = 20 * dpi
        height = 18 * dpi
        self.setIndentation(9 * dpi)
        self.setMinimumWidth(20 * dpi)
        self.setIconSize(QSize(width, height))
        self.setStyleSheet('height: {}px'.format(height))

    # ============================================================================================================
    # SETTINGS
    # ============================================================================================================

    def settings(self):
        """
        Returns the current state of the item as a dictionary
        :return: dict
        """

        settings = dict()

        vertical_scroll_bar = self.verticalScrollBar()
        horizontal_scroll_bar = self.horizontalScrollBar()
        settings['verticalScrollBar'] = {'value': vertical_scroll_bar.value()}
        settings['horizontalScrollBar'] = {'value': horizontal_scroll_bar.value()}

        for item in self.items():
            item_settings = item.settings()
            if item_settings:
                settings[item.path()] = item.settings()

        return settings

    def set_settings(self, settings_dict):
        """
        Sets the current state of the item from a dictionary
        :param settings_dict: dict
        """

        for path in sorted(list(settings_dict.keys())):
            setting_path = settings_dict.get(path, None)
            self.set_path_settings(path, setting_path)

        vertical_scrollbar_setting = settings_dict.get('verticalScrollBar', dict())
        value = vertical_scrollbar_setting.get('value', None)
        if value:
            self.verticalScrollBar().setValue(value)

        horizontal_scrollbar_setting = settings_dict.get('horizontalScrollBar', dict())
        value = horizontal_scrollbar_setting.get('value', None)
        if value:
            self.horizontalScrollBar().setValue(value)

        self.set_dpi(self.dpi())

    def set_path_settings(self, path, settings):
        """
        Set paths settings
        :param path: list(str)
        :param settings: dict
        """

        item = self.item_from_path(path)
        if item and settings:
            item.set_settings(settings)


class SidebarTreeItem(QTreeWidgetItem):

    _PIXMAP_CACHE = dict()

    def __init__(self, *args, **kwargs):
        super(SidebarTreeItem, self).__init__(*args, **kwargs)

        self._path = ''
        self._bold = None
        self._icon_visible = True
        self._icon_path = None
        self._icon_color = None
        self._text_color = None
        self._expanded_icon_path = None
        self._collapsed_icon_path = None

        self._settings = dict()

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def setSelected(self, flag):
        """
        Overrides base LibrarySideBarWidget setSelected function
        :param select: bool
        """

        super(SidebarTreeItem, self).setSelected(flag)
        if flag:
            self.set_expanded_parents(flag)

    def textColor(self):
        """
        Returns the foreground color of the item
        :return: QColor
        """

        clr = self.foreground(0).color()
        return color.Color.from_color(clr)

    def setTextColor(self, text_color):
        """
        Sets the foreground color to the given color
        :param text_color: variant, QColor or str
        """

        if isinstance(text_color, QColor):
            text_color = color.Color.from_color(text_color)
        elif python.is_string(text_color):
            text_color = color.Color.from_string(text_color)
        self._settings['textColor'] = text_color.to_string()
        brush = QBrush()
        brush.setColor(text_color)
        self.setForeground(0, brush)

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def path(self):
        """
        Returns item path
        :return: str
        """

        return self._path

    def set_path(self, path):
        """
        Sets the path for the item
        :param path: str
        """

        self._path = path

    def default_icon_path(self):
        """
        Returns the default icon path
        :return: str
        """

        return resources.get('icons', 'folder.svg')

    def icon_path(self):
        """
        Returns the icon path for the item
        :return: str
        """

        return self._icon_path or self.default_icon_path()

    def set_icon_path(self, path):
        """
        Sets the icon path for the item
        :param path: str
        """

        self._icon_path = path
        self.update_icon()

    def expanded_icon_path(self):
        """
        Returns the icon path to be shown when expanded
        :return: str
        """

        return self._expanded_icon_path or resources.get('icons', 'black', 'open_folder.png')

    def collapsed_icon_path(self):
        """
        Returns the icon path to be shown when collapsed
        :return: str
        """

        return self._collapsed_icon_path or resources.get('icons', 'black', 'folder.png') or self.default_icon_path()

    def is_icon_visible(self):
        """
        Returns whether or not the tree item icon is visible
        :return: bool
        """

        return self.treeWidget().icons_visible() and self._icon_visible

    def set_icon_visible(self, flag):
        """
        Sets whether or not tree item icon is visible
        :param flag: bool
        """

        self._icon_visible = flag

    def default_icon_color(self):
        """
        Returns the default icon color
        :return: str
        """

        palette = self.treeWidget().palette()
        default_color = palette.color(self.treeWidget().foregroundRole())
        default_color = color.Color.from_color(default_color).to_string()

        return str(default_color)

    def icon_color(self):
        """
        Returns the icon color
        :return: variant, QColor or None
        """

        return self._icon_color or self.default_icon_color()

    def set_icon_color(self, icon_color):
        """
        Sets the icon color
        :param icon_color: variant, QColor or str
        """

        if icon_color:
            if isinstance(icon_color, QColor):
                icon_color = color.Color.from_color(icon_color)
            elif python.is_string(icon_color):
                icon_color = color.Color.from_string(icon_color)
            self._icon_color = icon_color.to_string()
        else:
            self._icon_color = None

        self.update_icon()

    def update_icon(self):
        """
        Forces the icon to update
        """

        if not self.is_icon_visible():
            return

        if self.isExpanded():
            path = self.expanded_icon_path()
        else:
            path = self.collapsed_icon_path()
        item_pixmap = self._create_pixmap(path, self.icon_color())
        self.setIcon(0, item_pixmap)

    def update(self):
        """
        Updates current item
        """

        self.update_icon()

    def bold(self):
        """
        Returns whether item text is bold or not
        :return: bool
        """

        return self.font(0).bold()

    def set_bold(self, flag):
        """
        Sets whether item text is bold or not
        :param flag: bool
        """

        if flag:
            self._settings['bold'] = flag

        font = self.font(0)
        font.setBold(flag)
        self.setFont(0, font)

    def url(self):
        """
        Returns the url path
        :return: QUrl
        """

        return QUrl(self.path())

    def parents(self):
        """
        Returns all item parents
        :return: list(LibrarySidebarWidgetItem)
        """

        parents = list()
        parent = self.parent()
        if parent:
            parents.append(parent)
            while parent.parent():
                parent = parent.parent()
                parents.append(parent)

        return parents

    def set_expanded_parents(self, expanded):
        """
        Sets all the parents of the item to the value of expanded
        :param expanded: bool
        """

        parents = self.parents()
        for parent in parents:
            parent.setExpanded(expanded)

    def settings(self):
        """
        Returns the current state of the item as a dictionary
        :return: dict
        """

        settings = dict()

        is_selected = self.isSelected()
        if is_selected:
            settings['selected'] = is_selected
        is_expanded = self.isExpanded()
        if is_expanded:
            settings['expanded'] = is_expanded
        bold = self._settings.get('bold')
        if bold:
            settings['bold'] = bold
        text_color = self._settings.get('textColor')
        if text_color:
            settings['textColor'] = text_color

        return settings

    def set_settings(self, settings):
        """
        Sets the current state of the item from a dictionary
        :param settings: dict
        """

        text = settings.get('text')
        if text:
            self.setText(0, text)
        icon_path = settings.get('icon')
        if icon_path:
            self.set_icon_path(icon_path)
        icon_color = settings.get('color')
        if icon_color:
            self.set_icon_color(icon_color)
        is_selected = settings.get('selected')
        if is_selected is not None:
            self.setSelected(is_selected)
        is_expanded = settings.get('expanded')
        if is_expanded is not None and self.childCount() > 0:
            self.setExpanded(is_expanded)
            self.update_icon()
        bold = settings.get('bold')
        if bold is not None:
            self.set_bold(bold)
        text_color = settings.get('textColor')
        if text_color:
            self.setTextColor(text_color)

    # ============================================================================================================
    # INTERNAL
    # ============================================================================================================

    def _create_pixmap(self, path, color):
        """
        Internal function that creates a new item pixmap from the given path
        :param path: str
        :param color: str or QColor
        :return: QPixmap
        """

        if not path:
            return QPixmap()

        dpi = self.treeWidget().dpi()
        key = path + color + 'DPI-' + str(dpi)
        item_pixmap = self._PIXMAP_CACHE.get(key)
        if not item_pixmap:
            width = 20 * dpi
            height = 18 * dpi
            if '/' not in path and '\\' not in path:
                path = resources.get('icons', path)
            if not path or not os.path.exists(path):
                path = self.default_icon_path()
            pixmap2 = pixmap.Pixmap(path)
            pixmap2.set_color(color)
            pixmap2 = pixmap2.scaled(16 * dpi, 16 * dpi, Qt.KeepAspectRatio, Qt.SmoothTransformation)
            x = (width - pixmap2.width()) / 2
            y = (height - pixmap2.height()) / 2
            item_pixmap = QPixmap(QSize(width, height))
            item_pixmap.fill(Qt.transparent)
            painter = QPainter(item_pixmap)
            painter.drawPixmap(x, y, pixmap2)
            painter.end()
            self._PIXMAP_CACHE[key] = item_pixmap

        return item_pixmap



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains base item factory implementation for Maya
"""

import os

from tpDcc.libs.datalibrary.dccs.maya.data import mayaascii, mayabinary

from tpDcc.tools.datalibrary.core import factory
from tpDcc.tools.datalibrary.dccs.maya.widgets import save, load


class MayaItemsFactory(factory.BaseItemsFactory):
    def __init__(self, *args, **kwargs):
        super(MayaItemsFactory, self).__init__(*args, **kwargs)

    def _register_default_paths(self):
        super(MayaItemsFactory, self)._register_default_paths()

        self.register_path(
            os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'data'))

    def get_save_widget_class(self, data_instance):
        if data_instance in (mayaascii.MayaAsciiData, mayabinary.MayaBinaryData):
            return save.MayaSaveWidget

        return super(MayaItemsFactory, self).get_save_widget_class(data_instance)

    def get_load_widget_class(self, data_instance):
        if data_instance in (mayaascii.MayaAsciiData, mayabinary.MayaBinaryData):
            return load.MayaLoadWidget

        return super(MayaItemsFactory, self).get_load_widget_class(data_instance)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains status data library widget implementation
"""

from __future__ import print_function, division, absolute_import

from tpDcc.libs.qt.core import statusbar
from tpDcc.libs.qt.widgets import progressbar


class DataStatusWidget(statusbar.StatusWidget):

    DEFAULT_DISPLAY_TIME = 5000  # milliseconds -> 5 seconds

    def __init__(self, *args, **kwargs):
        super(DataStatusWidget, self).__init__(*args, **kwargs)

        self._progress_bar = progressbar.FrameProgressBar(self)
        self._progress_bar.setVisible(False)
        self.main_layout.addWidget(self._progress_bar)

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def progress_bar(self):
        """
        Returns the progress bar widget
        :return:  progressbar.ProgressBar
        """

        return self._progress_bar



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains library item tree view implementation
"""

from __future__ import print_function, division, absolute_import

import logging
import traceback

from Qt.QtCore import Qt, Signal, QPoint, QRect, QSize, QMimeData
from Qt.QtWidgets import QListView, QAbstractItemView, QRubberBand
from Qt.QtGui import QFont, QColor, QPixmap, QPalette, QPainter, QBrush, QDrag

from tpDcc.libs.qt.core import contexts as qt_contexts

from tpDcc.tools.datalibrary.core import consts
from tpDcc.tools.datalibrary.widgets import mixinview

LOGGER = logging.getLogger('tpDcc-tools-datalibrary')


class ViewerListView(mixinview.ViewerViewWidgetMixin, QListView):

    DEFAULT_DRAG_THRESHOLD = consts.LIST_DEFAULT_DRAG_THRESHOLD

    itemMoved = Signal(object)
    itemDropped = Signal(object)
    itemClicked = Signal(object)
    itemDoubleClicked = Signal(object)

    def __init__(self, *args, **kwargs):
        QListView.__init__(self, *args, **kwargs)
        mixinview.ViewerViewWidgetMixin.__init__(self)

        self.setSpacing(5)
        self.setMouseTracking(True)
        self.setSelectionRectVisible(True)
        self.setViewMode(QListView.IconMode)
        self.setResizeMode(QListView.Adjust)
        self.setSelectionMode(QListView.ExtendedSelection)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setAcceptDrops(True)
        self.setDragEnabled(True)
        self.setDragDropMode(QAbstractItemView.DragDrop)

        self._tree_widget = None
        self._rubber_band = None
        self._rubber_band_start_pos = None
        self._rubber_band_color = QColor(Qt.white)
        self._custom_sort_order = list()

        self._drag = None
        self._drag_start_pos = None
        self._drag_start_index = None
        self._drop_enabled = True

        self.clicked.connect(self._on_index_clicked)
        self.doubleClicked.connect(self._on_index_double_clicked)

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def startDrag(self, event):
        """
        Overrides bae QListView startDrag function
        :param event: QEvent
        """

        if not self.dragEnabled():
            return

        if self._drag_start_pos and hasattr(event, 'pos'):
            item = self.item_at(event.pos())
            if item and item.drag_enabled():
                self._drag_start_index = self.indexAt(event.pos())
                point = self._drag_start_pos - event.pos()
                dt = self.drag_threshold()
                if point.x() > dt or point.y() > dt or point.x() < -dt or point.y() < -dt:
                    items = self.selected_items()
                    mime_data = self.mime_data(items)
                    pixmap = self._drag_pixmap(item, items)
                    hotspot = QPoint(pixmap.width() * 0.5, pixmap.height() * 0.5)
                    self._drag = QDrag(self)
                    self._drag.setPixmap(pixmap)
                    self._drag.setHotSpot(hotspot)
                    self._drag.setMimeData(mime_data)
                    self._drag.start(Qt.MoveAction)

    def endDrag(self):
        """
        Function that ends current drag
        """

        self._drag_start_pos = None
        self._drag_start_index = None
        if self._drag:
            del self._drag
            self._drag = None

    def dragEnterEvent(self, event):
        """
        Overrides bae QListView dragEnterEvent function
        :param event: QDragEvent
        """

        mimedata = event.mimeData()
        if (mimedata.hasText() or mimedata.hasUrls()) and self.drop_enabled():
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        """
        Overrides bae QListView dragMoveEvent function
        :param event: QDragEvent
        """

        mimedata = event.mimeData()
        if (mimedata.hasText() or mimedata.hasUrls()) and self.drop_enabled():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        """
        Overrides bae QListView dropEvent function
        :param event: QDropEvent
        """

        item = self.item_at(event.pos())
        selected_items = self.selected_item()
        if selected_items and item:
            if self.tree_widget().is_sort_by_custom_order():
                self.move_items(selected_items, item)
            else:
                LOGGER.info('You can only re-order items when sorting by custom order')

        if item:
            item.drop_event(event)

        self.itemDropped.emit(event)

    # ============================================================================================================
    # OVERRIDES - MIXIN
    # ============================================================================================================

    def mousePressEvent(self, event):
        """
        Overrides base QListView mousePressEvent function
        :param event: QMouseEvent
        """

        item = self.item_at(event.pos())
        if not item:
            self.clearSelection()

        mixinview.ViewerViewWidgetMixin.mousePressEvent(self, event)
        if event.isAccepted():
            QListView.mousePressEvent(self, event)
            if item:
                # NOTE: This causes viewer tree widget selectionChanged signal to be emitted multiple times.
                # NOTE: This causes that item preview widgets are created twice when selecting an item in the viewer.
                # NOTE: For this reason, we block tree widgets signals before selecting the item
                with qt_contexts.block_signals(self.tree_widget()):
                    item.setSelected(True)

        self.endDrag()
        self._drag_start_pos = event.pos()

        is_left_button = self.mouse_press_button() == Qt.LeftButton
        is_item_draggable = item and item.drag_enabled()
        is_selection_empty = not self.selected_items()

        if is_left_button and (is_selection_empty or not is_item_draggable):
            self.rubber_band_start_event(event)

    def mouseMoveEvent(self, event):
        """
        Overrides base QListView mouseMoveEvent function
        :param event: QMouseEvent
        """

        if not self.is_dragging_items():
            is_left_button = self.mouse_press_button() == Qt.LeftButton
            if is_left_button and self.rubber_band().isHidden() and self.selected_items():
                self.startDrag(event)
            else:
                mixinview.ViewerViewWidgetMixin.mouseMoveEvent(self, event)
                QListView.mouseMoveEvent(self, event)
            if is_left_button:
                self.rubber_band_move_event(event)

    def mouseReleaseEvent(self, event):
        """
        Override base QListView mouseReleaseEvent function
        :param event: QMouseEvent
        """

        item = self.item_at(event.pos())
        items = self.selected_items()
        mixinview.ViewerViewWidgetMixin.mouseReleaseEvent(self, event)
        if item not in items:
            if event.button() != Qt.MidButton:
                QListView.mouseReleaseEvent(self, event)
        elif not items:
            QListView.mouseReleaseEvent(self, event)

        self.endDrag()
        self.rubber_band().hide()

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def scroll_to_item(self, item, pos=None):
        """
        Ensures that the item is visible
        :param item: LibraryItem
        :param pos: QPoint or None
        """

        index = self.index_from_item(item)
        pos = pos or QAbstractItemView.PositionAtCenter

        self.scrollTo(index, pos)

    # ============================================================================================================
    # TREE WIDGET
    # ============================================================================================================

    def tree_widget(self):
        """
        Return the tree widget that contains the items
        :return: LibraryTreeWidget
        """

        return self._tree_widget

    def set_tree_widget(self, tree_widget):
        """
        Set the tree widget that contains the items
        :param tree_widget: LibraryTreeWidget
        """

        self._tree_widget = tree_widget
        self.setModel(tree_widget.model())
        self.setSelectionModel(tree_widget.selectionModel())

    def items(self):
        """
        Return all the items
        :return: list(LibraryItem)
        """

        return self.tree_widget().items()

    def row_at(self, pos):
        """
        Returns the row for the given pos
        :param pos: QPoint
        :return:
        """

        return self.tree_widget().row_at(pos)

    def item_at(self, pos):
        """
        Returns a pointer to the item at the coordinates p
        The coordinates are relative to the tree widget's viewport
        :param pos: QPoint
        :return: LibraryItem
        """

        index = self.indexAt(pos)
        return self.item_from_index(index)

    def selected_item(self):
        """
        Returns the last selected non-hidden item
        :return: QTreeWidgetItem
        """

        return self.tree_widget().selected_item()

    def selected_items(self):
        """
        Returns a list of all selected non-hidden items
        :return: list(QTreeWidgetItem)
        """

        return self.tree_widget().selectedItems()

    def insert_item(self, row, item):
        """
        Inserts the item at row in the top level in the view
        :param row: int
        :param item: QTreeWidgetItem
        """

        self.tree_widget().insertTopLevelItem(row, item)

    def take_items(self, items):
        """
        Removes and returns the items from the view
        :param items: list(QTreeWidgetItem)
        :return: list(QTreeWidgetItem)
        """

        for item in items:
            row = self.tree_widget().indexOfTopLevelItem(item)
            self.tree_widget().takeTopLevelItem(row)

        return items

    def set_indexes_selected(self, indexes, value):
        """
        Set the selected state for the given indexes
        :param indexes: list(QModelIndex)
        :param value: bool
        """

        items = self.items_from_indexes(indexes)
        self.set_items_selected(items, value)

    def set_items_selected(self, items, value):
        """
        Sets the selected state for the given items
        :param items: list(LibraryItem)
        :param value: bool
        """

        with qt_contexts.block_signals(self.tree_widget()):
            try:
                for item in items:
                    item.setSelected(value)
            except Exception:
                LOGGER.error(str(traceback.format_exc()))

    def move_items(self, items, item_at):
        """
        Moves the given items to the position at the given row
        :param items: list(LibraryItem)
        :param item_at: LibraryItem
        """

        scroll_value = self.verticalScrollBar().value()
        self.tree_widget().move_items(items, item_at)
        self.itemMoved.emit(items[-1])
        self.verticalScrollBar().setValue(scroll_value)

    def index_from_item(self, item):
        """
        Returns QModelIndex associated with the given item
        :param item: LibraryItem
        :return: QModelIndex
        """

        return self.tree_widget().indexFromItem(item)

    def item_from_index(self, index):
        """
        Return a pointer to the LibraryItem associated with the given model index
        :param index: QModelIndex
        :return: LibraryItem
        """

        return self.tree_widget().itemFromIndex(index)

    def items_from_urls(self, urls):
        """
        Returns items from the given URL objects
        :param urls: list(QUrl)
        :return: DataItem
        """

        items = list()
        for url in urls:
            item = self.item_from_url(url)
            if item:
                items.append(item)

        return items

    def item_from_url(self, url):
        """
        Returns the item from the given url object
        :param url: QUrl
        :return: DataItem
        """

        return self.item_from_path(url.path())

    def items_from_paths(self, paths):
        """
        Returns the items from the given paths
        :param paths: list(str)
        :return: QUrl
        """

        items = list()
        for path in paths:
            item = self.item_from_path(path)
            if item:
                items.append(item)

        return items

    def item_from_path(self, path):
        """
        Returns the item from the given path
        :param path: str
        :return: DataItem
        """

        for item in self.items():
            item_path = item.url().path()
            if item_path and path == item_path:
                return item

        return None

    # ============================================================================================================
    # DRAG & DROP
    # ============================================================================================================

    def drop_enabled(self):
        """
        Returns whether drop functionality is enabled or not
        :return: bool
        """

        return self._drop_enabled

    def set_drop_enabled(self, flag):
        """
        Sets whether drop functionality is enabled or not
        :param flag: bool
        """

        self._drop_enabled = flag

    def drag_threshold(self):
        """
        Returns current drag threshold
        :return: float
        """

        return self.DEFAULT_DRAG_THRESHOLD

    def is_dragging_items(self):
        """
        Returns whether the user is currently dragging items or not
        :return: bool
        """

        return bool(self._drag)

    def mime_data(self, items):
        """
        Returns drag mime data
        :param items: list(LibraryItem)
        :return: QMimeData
        """

        mimedata = QMimeData()
        urls = [item.url() for item in items]
        text = '\n'.join([item.mime_text() for item in items])
        mimedata.setUrls(urls)
        mimedata.setText(text)

        return mimedata

    # ============================================================================================================
    # RUBBER BAND
    # ============================================================================================================

    def create_rubber_band(self):
        """
        Creates a new instance of the selection rubber band
        :return: QRubberBand
        """

        rubber_band = QRubberBand(QRubberBand.Rectangle, self)
        palette = QPalette()
        color = self.rubber_band_color()
        palette.setBrush(QPalette.Highlight, QBrush(color))
        rubber_band.setPalette(palette)

        return rubber_band

    def rubber_band(self):
        """
        Retursn the selection rubber band for this widget
        :return: QRubberBand
        """

        if not self._rubber_band:
            self.setSelectionRectVisible(False)
            self._rubber_band = self.create_rubber_band()

        return self._rubber_band

    def rubber_band_color(self):
        """
        Returns the rubber band color for this widget
        :return: QColor
        """

        return self._rubber_band_color

    def set_rubber_band_color(self, color):
        """
        Sets the color for the rubber band
        :param color: QColor
        """

        self._rubber_band = None
        self._rubber_band_color = color

    def rubber_band_start_event(self, event):
        """
        Triggered when the user presses an empty area
        :param event: QMouseEvent
        """

        self._rubber_band_start_pos = event.pos()
        rect = QRect(self._rubber_band_start_pos, QSize())
        rubber_band = self.rubber_band()
        rubber_band.setGeometry(rect)
        rubber_band.show()

    def rubber_band_move_event(self, event):
        """
        Triggered when the user moves the mouse over the current viewport
        :param event: QMouseEvent
        """

        if self.rubber_band() and self._rubber_band_start_pos:
            rect = QRect(self._rubber_band_start_pos, event.pos())
            rect = rect.normalized()
            self.rubber_band().setGeometry(rect)

    # ============================================================================================================
    # INTERNAL
    # ============================================================================================================

    def _drag_pixmap(self, item, items):
        """
        Internal function that shows the pixmap for the given item during drag operation
        :param item: LibraryItem
        :param items: list(LibraryItem)
        :return: QPixmap
        """

        rect = self.visualRect(self.index_from_item(item))
        pixmap = QPixmap()
        pixmap = pixmap.grabWidget(self, rect)
        if len(items) > 1:
            custom_width = 35
            custom_padding = 5
            custom_text = str(len(items))
            custom_x = pixmap.rect().center().x() - float(custom_width * 0.5)
            custom_y = pixmap.rect().top() + custom_padding
            custom_rect = QRect(custom_x, custom_y, custom_width, custom_width)
            painter = QPainter(pixmap)
            painter.setRenderHint(QPainter.Antialiasing)
            painter.setPen(Qt.NoPen)
            painter.setBrush(self.viewer().background_selected_color())
            painter.drawEllipse(custom_rect.center(), float(custom_width * 0.5), float(custom_width * 0.5))
            font = QFont('Serif', 12, QFont.Light)
            painter.setFont(font)
            painter.setPen(self.viewer().text_selected_color())
            painter.drawText(custom_rect, Qt.AlignCenter, str(custom_text))

        return pixmap

    # ============================================================================================================
    # CALLBACKS
    # ============================================================================================================

    def _on_index_clicked(self, index):
        """
        Callback function that is called when the user clicks on an item
        :param index: QModelIndex
        """

        item = self.item_from_index(index)
        item.clicked()
        self.set_items_selected([item], True)
        self.itemClicked.emit(item)

    def _on_index_double_clicked(self, index):
        """
        Callback function that is called when the user double clicks on an item
        :param index: QModelIndex
        """

        item = self.item_from_index(index)
        self.set_items_selected([item], True)
        item.double_clicked()
        self.itemDoubleClicked.emit(item)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains library data item widget implementation
"""

from __future__ import print_function, division, absolute_import

import os
import logging
import traceback

from Qt.QtCore import Qt, QUrl
from Qt.QtWidgets import QApplication, QAction, QDialogButtonBox, QFileDialog, QMessageBox

from tpDcc.managers import resources
from tpDcc.libs.python import path as path_utils
from tpDcc.libs.qt.widgets import messagebox
from tpDcc.libs.datalibrary.core import exceptions

from tpDcc.tools.datalibrary.core.views import item

LOGGER = logging.getLogger('tpDcc-libs-datalibrary')


class DataItemView(item.ItemView):

    SAVE_WIDGET_CLASS = None
    LOAD_WIDGET_CLASS = None

    def __init__(self, data_item, library_window=None):
        super(DataItemView, self).__init__(data_item=data_item)

        self._modal = None
        self._library_window = None
        self._read_only = False
        self._ignore_exists_dialog = False

        if library_window:
            self.set_library_window(library_window)

        # self._item.metaDataChanged.connect(self._on_metadata_updated)
        # self._item.pathCopiedToClipboard.connect(self._on_path_copied_to_clipboard)
        # self._item.saving.connect(self._on_before_save_item)
        # self._item.saved.connect(self._on_item_saved)
        # self._item.copied.connect(self._on_item_copied)
        # self._item.renamed.connect(self._on_item_renamed)
        # self._item.deleted.connect(self._on_item_deleted)

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def thumbnail_path(self):
        """
        Return the thumbnail path for the item on disk
        :return: str
        """

        thumb_name = self.item.get_thumb_name() if self.item else None
        if not thumb_name:
            return super(DataItemView, self).thumbnail_path()

        thumb_path = self.item.get_thumb_path(thumb_name) if self.item else None
        if not thumb_path or not os.path.isfile(thumb_path):
            return self._default_thumbnail_path

        return thumb_path

    def mime_text(self):
        """
        Returns the mime text for drag and drop
        :return: str
        """

        # if self.path():
        #     file_path = path_utils.clean_path(os.path.join(self.path(), self.name()))
        #     if not os.path.isfile(file_path):
        #         file_path = self.path()
        #     return file_path

        return self.path()

    def url(self):
        """
        Used by the mime data when dragging/droping the item
        :return: Qurl
        """

        # if self.path():
        #     file_path = path_utils.clean_path(os.path.join(self.path(), self.name()))
        #     if not os.path.isfile(file_path):
        #         file_path = self.path()
        #     return QUrl('file:///{}'.format(file_path))

        return QUrl('file:///{}'.format(self.path()))

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def name(self):
        """
        Returns name of the item
        :return: str
        """

        return self._item.name()

    def path(self):
        """
        Returns the path for the item
        :return: str
        """

        return self._item.format_identifier()

    def library(self):
        """
        Returns library model for the item
        :return: DataLibrary
        """

        if not self._item.library and self.library_window():
            return self.library_window().library()

        return self._item.library

    def library_window(self):
        """
        Returns the library widget containing the item
        :return: LibraryWindow
        """

        return self._library_window

    def set_library_window(self, library_window):
        """
        Sets the library widget containing the item
        :param library_window: LibraryWindow
        """

        self._library_window = library_window

    def is_locked(self):
        """
        Returns whether or not this item is locked
        :return: bool
        """

        locked = False
        if self.library_window():
            locked = self.library_window().is_locked()

        return locked

    def is_read_only(self):
        """
        Returns whether or not this item is read only
        :return: bool
        """

        if self.is_locked():
            return True

        return self._read_only

    def set_read_only(self, flag):
        """
        Sets whether or not this item is read only
        :param flag: bool
        """

        self._read_only = flag

    def is_deletable(self):
        """
        Returns whether or not this item is deletable
        :return: bool
        """

        if self.is_locked():
            return False

        return self.item.ENABLE_DELETE

    def select_folder(self):
        """
        Select the folder in the library widget
        """

        if not self.library_window():
            return

        item_path = '/'.join(path_utils.normalize_path(self.path()).split('/')[:-1])
        self.library_window().select_folder_path(item_path)

    # ============================================================================================================
    # PREVIEW WIDGET
    # ============================================================================================================

    def preview_widget(self):
        """
        Returns the widget to be shown when the user clicks on the item
        :return: QWidget or None
        """

        widget = None

        load_widget_class = self.library_window().factory.get_load_widget_class(self.item.__class__)
        if load_widget_class:
            widget = load_widget_class(item_view=self, client=self.library_window().client)

        return widget

    def show_preview_widget(self, library_window):
        """
        Shows the preview widget for the item instance
        :param library_window: LibraryWindow
        """

        widget = self.preview_widget()
        library_window.set_preview_widget(widget)

    # ============================================================================================================
    # DIALOGS
    # ============================================================================================================

    def show_toast_message(self, text):
        """
        Function that shows the toast widget with the given text
        :param text: str
        """

        if self.library_window():
            self.library_window().show_toast_message(text)

    def show_error_dialog(self, title, text):
        """
        Function that shows an error dialog to the user
        :param title: str
        :param text: str
        :return: QMessageBox.StandardButton or None
        """

        if self.library_window():
            self.library_window().show_error_message(text)

        button = None
        if not self._modal:
            self._modal = True
            try:
                button = messagebox.MessageBox.critical(self.library_window(), title, text)
            finally:
                self._modal = False

        return button

    def show_exception_dialog(self, title, error, exception):
        """
        Function that shows a question dialog to the user
        :param title: str
        :param error: str
        :param exception: str
        """

        LOGGER.exception(exception)
        return self.show_error_dialog(title, error)

    def show_question_dialog(self, title, text):
        """
        Function that shows a question dialog to the user
        :param title: str
        :param text: str
        :return: QMessageBox.StandardButton
        """

        return messagebox.MessageBox.question(self.library_window(), title, text)

    def show_rename_dialog(self, parent=None):
        """
        Shows the rename dialog
        :param parent: QWidget
        """

        select = False
        if self.library_window():
            parent = parent or self.library_window()
            select = self.library_window().selected_folder_path() == self.path()

        name, btn = messagebox.MessageBox.input(
            parent, 'Rename item', 'Rename the current item to:', input_text=self.name())
        if btn == QDialogButtonBox.Ok:
            try:
                rename_function = self.item.functionality().get('rename')
                if not rename_function:
                    self.show_error_dialog('Rename Error', 'Item "{}" does not supports rename operation'.format(self))
                    return
                rename_function(name)
                if select:
                    self.library_window().select_folder_path(self.path())
            except Exception as exc:
                self.show_exception_dialog('Rename Error', exc, traceback.format_exc())
                raise

        return btn

    def show_move_dialog(self, parent=None):
        """
        Shows the move to browser dialog
        :param parent: QWidget
        """

        path = os.path.dirname(self.path())
        target = QFileDialog.getExistingDirectory(parent, 'Move To ...', path)
        if target:
            try:
                move_function = self.item.functionality().get('move')
                if not move_function:
                    self.show_error_dialog('Delete Error', 'Item "{}" does not supports move operation'.format(self))
                    return
                source = self.item.format_identifier()
                if os.path.isdir(source):
                    target = path_utils.join_path(target, os.path.basename(source))
                move_function(target)
            except Exception as exc:
                self.show_exception_dialog('Move Error', exc, traceback.format_exc())
                raise

    def show_delete_dialog(self, dependencies=False):
        """
        Shows the delete item dialog
        """

        dialog_message = 'Are you sure you want to delete this item?' if not \
            dependencies else 'Are you sure you want to delete this item and all its dependencies?'

        button = self.show_question_dialog('Delete Item', dialog_message)
        if button == QDialogButtonBox.Yes:
            try:
                if dependencies:
                    delete_function = self.item.functionality().get('delete_with_dependencies')
                else:
                    delete_function = self.item.functionality().get('delete')
                if not delete_function:
                    error_message = 'Item "{}" does not supports delete operation' if \
                        not dependencies else 'Item "{}" does not supports delete with dependencies operation'
                    self.show_error_dialog('Delete Error', error_message.format(self))
                    return
                delete_function()

            except Exception as exc:
                self.show_exception_dialog('Delete Error', exc, traceback.format_exc())
                raise

        self.library_window().set_preview_widget_from_item(None)

    def show_already_existing_dialog(self):
        """
        Shows a warning dialog if the item already exists on save
        """

        if not self.library_window():
            raise exceptions.ItemSaveError('Item already exists!')

        buttons = QDialogButtonBox.Yes | QDialogButtonBox.Cancel
        try:
            QApplication.setOverrideCursor(Qt.ArrowCursor)
            button = self.library_window().show_question_dialog(
                'Item already exists',
                'Would you like to move the existing item "{}" to trash?'.format(os.path.basename(self.path())), buttons
            )
        finally:
            QApplication.restoreOverrideCursor()

        if button == QDialogButtonBox.Yes:
            self._move_to_trash()
        elif button == QMessageBox.Cancel:
            return button
        else:
            raise exceptions.ItemSaveError('You cannot save over an existing item.')

        return button

    # ============================================================================================================
    # CONTEXTUAL MENUS
    # ============================================================================================================

    def context_edit_menu(self, menu, items=None):
        """
        This function is called when the user opens context menu
        The given menu is shown as a submenu of the main context menu
        This function can be override to create custom context menus in LibraryItems
        :param menu: QMenu
        :param items: list(LibraryItem)
        """

        rename_action = QAction(resources.icon('rename'), 'Rename', menu)
        rename_action.triggered.connect(self._on_show_rename_dialog)
        menu.addAction(rename_action)

        move_to_action = QAction(resources.icon('move'), 'Move to', menu)
        move_to_action.triggered.connect(self._on_move_dialog)
        menu.addAction(move_to_action)

        copy_path_action = QAction(resources.icon('copy'), 'Copy Path', menu)
        copy_path_action.triggered.connect(self._on_copy_path)
        menu.addAction(copy_path_action)

        if self.library_window():
            select_folder_action = QAction(resources.icon('select'), 'Select Folder', menu)
            select_folder_action.triggered.connect(self._on_select_folder)
            menu.addAction(select_folder_action)

        show_in_folder_action = QAction(resources.icon('folder'), 'Show in Folder', menu)
        show_in_folder_action.triggered.connect(self._on_show_in_folder)
        menu.addAction(show_in_folder_action)

        if self.is_deletable():
            delete_action = QAction(resources.icon('delete'), 'Delete', menu)
            delete_action.triggered.connect(self._on_show_delete_dialog)
            delete_with_dependencies_action = QAction(resources.icon('delete'), 'Delete with Dependencies', menu)
            delete_with_dependencies_action.triggered.connect(self._on_show_delete_with_dependencies_dialog)
            menu.addSeparator()
            menu.addAction(delete_action)
            menu.addAction(delete_with_dependencies_action)

        self.create_overwrite_menu(menu)

    def create_overwrite_menu(self, menu):
        """
        Creates a menu or action to trigger the overwrite functionality
        :param menu: QMenu
        """

        if self.is_read_only():
            return

        menu.addSeparator()
        overwrite_action = QAction(resources.icon('replace'), 'Overwrite', menu)
        overwrite_action.triggered.connect(self._on_overwrite)
        menu.addAction(overwrite_action)

    # ============================================================================================================
    # INTERNAL
    # ============================================================================================================

    def _move_to_trash(self):
        """
        Internal function that moves current item to trash
        """

        path = self.path()
        library = self.library()
        item = DataItemView(path, library=library)
        self.library_window().move_items_to_trash([item])

    # ============================================================================================================
    # CALLBACKS
    # ============================================================================================================

    def _on_show_rename_dialog(self):
        """
        Internal callback function that is triggered when Rename action is executed
        """

        self.show_rename_dialog()

    def _on_move_dialog(self):
        """
        Internal callback function that is triggered when Move action is executed
        """

        self.show_move_dialog()

    def _on_copy_path(self):
        """
        Internal callback function that is triggered when Copy Path action is executed
        """

        copy_path_to_clipboard_function = self.item.functionality().get('copy_path_to_clipboard')
        if not copy_path_to_clipboard_function:
            self.show_error_dialog(
                'Copy Path Error', 'Item "{}" does not supports copy path to clipboard operation'.format(self))
            return

        copy_path_to_clipboard_function()

    def _on_select_folder(self):
        """
        Internal callback function that is called when Select Folder action is clicked
        """

        self.select_folder()

    def _on_show_in_folder(self):
        """
        Internal callback function that is called when Show in Folder action is clicked
        """

        show_in_explorer_function = self.item.functionality().get('show_in_explorer')
        if not show_in_explorer_function:
            self.show_error_dialog(
                'Show in Explorer Error', 'Item "{}" does not supports show in explorer operation'.format(self))
            return

        show_in_explorer_function()

    def _on_show_delete_dialog(self):
        """
        Internal callback function that is called when Delete action is clicked
        """

        self.show_delete_dialog()

    def _on_show_delete_with_dependencies_dialog(self):
        """
        Internal callback function that is called when Delte with Dependencies action is clicked
        """

        self.show_delete_dialog(dependencies=True)

    def _on_overwrite(self):
        """
        Internal callback function that is called when Overwrite action is clicked
        """

        self._ignore_exists_dialog = True
        widget = self.show_save_widget(self.library_window(), item_view=self)

    def _on_metadata_updated(self, metadata_dict):
        """
        Internal callback function that is called when metadata is updated
        :param metadata_dict: dict
        """

        pass

    def _on_path_copied_to_clipboard(self):
        """
        Internal callback function that is called when an item path is copied to clipboard
        """

        if self.library_window():
            self.library_window().show_success_message('Path copied to clipboard successfully!')

    def _on_before_save_item(self, save_path):
        if os.path.exists(save_path):
            if self._ignore_exists_dialog:
                self._move_to_trash()
            else:
                res = self.show_already_existing_dialog()
                if res == QMessageBox.Cancel:
                    self.item.cancel_save()
                    return False

        return True

    def _on_item_saved(self):
        if self.library_window():
            self.library_window().sync()
            self.library_window().select_items([self])

    def _on_item_copied(self):
        if self.library_window():
            self.library_window().refresh()

    def _on_item_renamed(self):
        if self.library_window():
            self.library_window().sync()

    def _on_item_deleted(self):
        if self.library_window():
            self.library_window().sync()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains data library sort by menu widget implementation
"""

from __future__ import print_function, division, absolute_import

from functools import partial

from Qt.QtWidgets import QMenu
from Qt.QtGui import QCursor

from tpDcc.libs.qt.widgets import action


class SortByMenu(QMenu, object):
    def __init__(self, *args, **kwargs):
        super(SortByMenu, self).__init__(*args, **kwargs)

        self._library = None

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def show(self, point=None):
        """
        Overrides base QMenu show function
        :param point: QPoint
        """

        self.clear()

        sort_by = self.library().sort_by()
        if sort_by:
            current_field = self.library().sort_by()[0].split(':')[0]
            current_order = 'dsc' if 'dsc' in self.library().sort_by()[0] else 'asc'
        else:
            current_field = ''
            current_order = ''

        separator_action = action.SeparatorAction('Sort By', self)
        self.addAction(separator_action)

        fields = self.library().fields()
        for field in fields:
            if not field.get('sortable'):
                continue
            name = field.get('name')
            field_action = self.addAction(name.title())
            field_action.setCheckable(True)
            field_action.setChecked(bool(current_field == name))
            field_callback = partial(self.set_sort_by, name, current_order)
            field_action.triggered.connect(field_callback)

        separator_action = action.SeparatorAction('Sort Order', self)
        self.addAction(separator_action)

        ascending_action = self.addAction('Ascending')
        ascending_action.setCheckable(True)
        ascending_action.setChecked(bool(current_order == 'asc'))
        ascending_callback = partial(self.set_sort_by, current_field, 'asc')
        ascending_action.triggered.connect(ascending_callback)

        descending_action = self.addAction('Descending')
        descending_action.setCheckable(True)
        descending_action.setChecked(bool(current_order == 'dsc'))
        descending_callback = partial(self.set_sort_by, current_field, 'dsc')
        descending_action.triggered.connect(descending_callback)

        poinrt = point or QCursor.pos()

        self.exec_(point)

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def library(self):
        """
        Returns the library model for the menu
        :return: Library
        """

        return self._library

    def set_library(self, library):
        """
        Set the library model for the menu
        :param library: Library
        """

        self._library = library

    def set_sort_by(self, sort_name, sort_order):
        """
        Sets the sort by value for the library
        :param sort_name: str
        :param sort_order: str
        """

        if sort_name == 'Custom Order':
            sort_order = 'asc'

        value = sort_name + ':' + sort_order
        self.library().set_sort_by([value])
        self.library().search()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains utils functions used by tpDcc-tools-datalibrary
"""

from __future__ import print_function, division, absolute_import

import os
import json
import locale
import shutil
import logging
import tempfile
from collections import OrderedDict, Mapping

from tpDcc.managers import configs
from tpDcc.libs.python import osplatform, path as path_utils

LOGGER = logging.getLogger('tpDcc-tools-datalibrary')


def absolute_path(data, start):
    """
    Returns an absolute version of all the paths in data using the start path
    :param data: str
    :param start: str
    :return: str
    """

    rel_path1 = path_utils.normalize_path(os.path.dirname(start))
    rel_path2 = path_utils.normalize_path(os.path.dirname(rel_path1))
    rel_path3 = path_utils.normalize_path(os.path.dirname(rel_path2))

    if not rel_path1.endswith("/"):
        rel_path1 += "/"

    if not rel_path2.endswith("/"):
        rel_path2 += "/"

    if not rel_path3.endswith("/"):
        rel_path3 += "/"

    data = data.replace('../../../', rel_path3)
    data = data.replace('../../', rel_path2)
    data = data.replace('../', rel_path1)

    return data


def update(data, other):
    """
    Update teh value of a nested dictionary of varying depth
    :param data: dict
    :param other: dict
    :return: dict
    """

    for key, value in other.items():
        if isinstance(value, Mapping):
            data[key] = update(data.get(key, {}), value)
        else:
            data[key] = value

    return data


def read(path):
    """
    Returns the contents of the given file
    :param path: str
    :return: str
    """

    data = ''
    path = path_utils.normalize_path(path)
    if os.path.isfile(path):
        with open(path) as f:
            data = f.read() or data
    data = absolute_path(data, path)

    return data


def write(path, data):
    """
    Writes the given data to the given file on disk
    :param path: str
    :param data: str
    """

    path = path_utils.normalize_path(path)
    data = path_utils.get_relative_path(data, path)

    tmp = path + '.tmp'
    bak = path + '.bak'

    dirname = os.path.dirname(path)
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    if os.path.exists(tmp):
        try:
            os.remove(tmp)
        except Exception:
            pass
        if os.path.exists(tmp):
            msg = 'The path is locked for writing and cannot be accessed {}'.format(tmp)
            raise IOError(msg)

    try:
        with open(tmp, 'w') as f:
            f.write(data)
            f.flush()

        if os.path.exists(bak):
            os.remove(bak)
        if os.path.exists(path):
            os.rename(path, bak)
        if os.path.exists(tmp):
            os.rename(tmp, path)
        if os.path.exists(path) and os.path.exists(bak):
            os.remove(bak)
    except Exception:
        if os.path.exists(tmp):
            try:
                os.remove(tmp)
            except Exception:
                pass
        if not os.path.exists(path) and os.path.exists(bak):
            os.rename(bak, path)

        raise


def update_json(path, data):
    """
    Update a JSON file with the given data
    :param path: str
    :param data: dict
    """

    data_ = read_json(path)
    data_ = update(data_, data)
    save_json(path, data_)


def read_json(path):
    """
    Reads the given JSON file and deserialize it to a Python object
    :param path: str
    :return: dict
    """

    path = path_utils.normalize_path(path)
    data = read(path) or '{}'
    data = json.loads(data)

    return data


def save_json(path, data):
    """
    Serialize given tdata to a JSON string and write it to the given path
    :param path: str
    :param data: dict
    """

    path = path_utils.normalize_path(path)
    data = OrderedDict(sorted(data.items(), key=lambda t: t[0]))
    data = json.dumps(data, indent=4)
    write(path, data)


def replace_json(path, old, new, count=-1):
    """
    Replaces the old value with the new value in the given JSON file
    :param path: str
    :param old: str
    :param new: str
    :param count: int
    :return: dict
    """

    old = str(old.encode("unicode_escape"))
    new = str(new.encode("unicode_escape"))

    data = read(path) or "{}"
    data = data.replace(old, new, count)
    data = json.loads(data)

    save_json(path, data)

    return data


def format_path(format_string, path='', **kwargs):
    """
    Resolves given path by replacing necessary info with proper data
    :param format_string: str
    :param path: str
    :param kwargs:
    :return: str
    """

    dirname, name, extension = path_utils.split_path(path)
    encoding = locale.getpreferredencoding()

    temp = tempfile.gettempdir()
    if temp:
        try:
            temp = temp.decode(encoding)
        except Exception:
            pass

    username = osplatform.get_user(lower=True)
    if username:
        try:
            username = username.decode(encoding)
        except Exception:
            pass

    local = os.getenv('APPDATA') or os.getenv('HOME')
    if local:
        try:
            local = local.decode(encoding)
        except Exception:
            pass

    kwargs.update(os.environ)

    format_dict = {
        'name': name,
        'path': path,
        'root': path,
        'user': username,
        'temp': temp,
        'home': local,
        'local': local,
        'dirname': dirname,
        'extension': extension
    }
    kwargs.update(format_dict)

    resolve_string = path_utils.normalize_path(format_string).format(**kwargs)

    return path_utils.clean_path(resolve_string)


def copy_path(source, target, force=False):
    """
    Makes a copy of the given source path to the given destination path
    :param source: str
    :param target: str
    :param force: bool
    :return: str
    """

    dirname = os.path.dirname(source)
    if '/' not in target:
        target = os.path.join(dirname, target)

    source = path_utils.normalize_path(source)
    target = path_utils.normalize_path(target)

    if source == target:
        raise IOError('The source path and destination path are the same: {}'.format(source))
    if not force and os.path.exists(target):
        raise IOError('Cannot copy over an existing path: {}'.format(target))

    if force and os.path.exists(target):
        if os.path.isdir(target):
            shutil.rmtree(target)
        else:
            os.remove(target)

    # Ensure target directory exists
    target_directory = os.path.dirname(target)
    if not os.path.exists(target_directory):
        os.makedirs(target_directory)

    if os.path.isfile(source):
        shutil.copy(source, target)
    else:
        shutil.copytree(source, target)

    return target


def temp_path(*args):
    """
    Returns the temporal directory based on library settings
    :param args:
    """

    datalib_config = configs.get_library_config('tpDcc-libs-datalibrary')
    temp_path = datalib_config.get('temp_path')
    if not temp_path:
        temp_path = tempfile.mkdtemp()

    temp_path = path_utils.normalize_path(os.path.join(format_path(temp_path), *args))

    return temp_path


def settings_path():
    """
    Returns library settings file path
    :return: str
    """

    datalib_config = configs.get_tool_config('tpDcc-tools-datalibrary')
    settings_file_path = datalib_config.get('settings_path')
    settings_file_path = format_path(settings_file_path)

    return settings_file_path


def read_settings(settings_file_path):
    """
    Returns all user settings
    :return: dict
    """

    path = settings_file_path or settings_path()
    data = dict()
    try:
        data = read_json(path)
    except Exception:
        LOGGER.exception('Cannot read settings from "{}"'.format(path))

    return data


def update_settings(data, settings_file_path=None):
    """
    Updates the settings with the given data
    :param data: dict
    :param settings_file_path: str or None
    """

    settings = read_settings(settings_file_path)
    update(settings, data)
    save_settings(settings)


def save_settings(data, settings_file_path=None):
    """
    Saves the given data to the settings path
    :param data: dict
    :param settings_file_path: str or None
    """

    path = settings_file_path or settings_path()
    try:
        save_json(path, data)
    except Exception:
        LOGGER.exception('Cannot save settings to "{}"'.format(path))



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains data library viewer widget implementation
"""

from __future__ import print_function, division, absolute_import

import logging
from functools import partial

from Qt.QtCore import Qt, Signal, QSize, QEvent
from Qt.QtWidgets import QApplication, QStyledItemDelegate, QAbstractItemView, QMenu, QAction
from Qt.QtGui import QCursor, QColor

from tpDcc.managers import resources
from tpDcc.libs.qt.core import base, qtutils, contexts as qt_contexts
from tpDcc.libs.qt.widgets import layouts, toast, action

from tpDcc.tools.datalibrary.core import consts
from tpDcc.tools.datalibrary.core.views import item
from tpDcc.tools.datalibrary.data import group
from tpDcc.tools.datalibrary.widgets import listview, treeview

LOGGER = logging.getLogger('tpDcc-libs-datalibrary')


class DataViewerDelegate(QStyledItemDelegate, object):
    """
    Class that defines visual style in LibraryViewer of LibraryItems
    """

    def __init__(self):
        super(DataViewerDelegate, self).__init__()

        self._viewer = None

    def sizeHint(self, option, index):
        """
        Overrides base QStyledItemDelegate sizeHint function
        Return the size for the given id¡ndex
        :param option: QStylOptionViewItem
        :param index: QModelIndex
        :return: QSize
        """

        item = self.viewer().item_from_index(index)
        if isinstance(item, group.GroupDataItemView):
            return item.sizeHint()

        return self.viewer().item_size_hint()

    def paint(self, painter, option, index):
        """
        Overrides base QStyledItemDelegate paint function
        Paint performs low-level painting for the given model index
        :param painter: QPainter
        :param option: QStyleOptionViewItem
        :param index: QModelIndex
        """

        item = self.viewer().item_from_index(index)
        item.paint(painter, option, index)

    def viewer(self):
        """
        Returns LibraryViewer object associated to this delegate
        :return: LibraryViewer
        """

        return self._viewer

    def set_viewer(self, viewer):
        """
        Set LibraryViewer associated to this delegate
        :param viewer: LibraryViewer
        """

        self._viewer = viewer


class DataViewer(base.BaseWidget):

    IconMode = consts.DEFAULT_ICON_MODE
    TableMode = consts.DEFAULT_TABLE_MODE

    DEFAULT_PADDING = consts.VIEWER_DEFAULT_PADDING
    DEFAULT_ZOOM_AMOUNT = consts.VIEWER_DEFAULT_ZOOM_AMOUNT
    DEFAULT_TEXT_HEIGHT = consts.VIEWER_DEFAULT_TEXT_HEIGHT
    DEFAULT_WHEEL_SCROLL_STEP = consts.VIEWER_DEFAULT_WHEEL_SCROLL_STEP
    DEFAULT_MIN_SPACING = consts.VIEWER_DEFAULT_MIN_SPACING
    DEFAULT_MAX_SPACING = consts.VIEWER_DEFAULT_MAX_SPACING
    DEFAULT_MIN_LIST_SIZE = consts.VIEWER_DEFAULT_MIN_LIST_SIZE
    DEFAULT_MIN_ICON_SIZE = consts.VIEWER_DEFAULT_MIN_ICON_SIZE

    DEFAULT_TEXT_COLOR = consts.VIEWER_DEFAULT_TEXT_COLOR
    DEFAULT_SELECTED_TEXT_COLOR = consts.VIEWER_DEFAULT_SELECTED_TEXT_COLOR
    DEFAULT_BACKGROUND_COLOR = consts.VIEWER_DEFAULT_BACKGROUND_COLOR
    DEFAULT_BACKGROUND_HOVER_COLOR = consts.VIEWER_DEFAULT_BACKGROUND_HOVER_COLOR
    DEFAULT_BACKGROUND_SELECTED_COLOR = consts.VIEWER_DEFAULT_BACKGROUND_SELECTED_COLOR

    LABEL_DISPLAY_OPTION = item.LabelDisplayOption.Under

    TREE_WIDGET_CLASS = treeview.ViewerTreeView
    LIST_VIEW_CLASS = listview.ViewerListView
    DELEGATE_CLASS = DataViewerDelegate

    itemClicked = Signal(object)
    itemDoubleClicked = Signal(object)
    zoomChanged = Signal(object)
    spacingChanged = Signal(object)
    groupClicked = Signal(object)
    keyPressed = Signal(object)

    def __init__(self, library_window):

        self._dpi = 1
        self._padding = self.DEFAULT_PADDING
        self._library = None
        self._library_window = library_window

        self._tree_widget = None
        self._list_view = None
        self._delegate = None
        self._is_item_text_visible = True
        self._toast_enabled = True

        self._zoom_amount = self.DEFAULT_ZOOM_AMOUNT
        self._icon_size = QSize(self._zoom_amount, self._zoom_amount)
        self._item_size_hint = QSize(self._zoom_amount, self._zoom_amount)
        self._label_display_option = self.LABEL_DISPLAY_OPTION
        self._text_color = self.DEFAULT_TEXT_COLOR
        self._text_selected_color = self.DEFAULT_SELECTED_TEXT_COLOR
        self._background_color = self.DEFAULT_BACKGROUND_COLOR
        self._background_hover_color = self.DEFAULT_BACKGROUND_HOVER_COLOR
        self._background_selected_color = self.DEFAULT_BACKGROUND_SELECTED_COLOR

        super(DataViewer, self).__init__(parent=library_window)

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def get_main_layout(self):
        main_layout = layouts.HorizontalLayout()

        return main_layout

    def ui(self):
        super(DataViewer, self).ui()

        self._tree_widget = self.TREE_WIDGET_CLASS(self)
        self._list_view = self.LIST_VIEW_CLASS(self)
        self._list_view.set_tree_widget(self._tree_widget)

        self._toast_widget = toast.ToastWidget(self)
        self._toast_widget.setVisible(False)

        self._delegate = self.DELEGATE_CLASS()
        self._delegate.set_viewer(self)
        self._tree_widget.setItemDelegate(self._delegate)
        self._list_view.setItemDelegate(self._delegate)

        self.main_layout.addWidget(self._tree_widget)
        self.main_layout.addWidget(self._list_view)

        self._tree_widget.installEventFilter(self)
        self._list_view.installEventFilter(self)

        self.itemMoved = self._list_view.itemMoved
        self.itemDropped = self._list_view.itemDropped
        self.itemSelectionChanged = self._tree_widget.itemSelectionChanged

    def setup_signals(self):
        header = self.tree_widget().header()
        header.sortIndicatorChanged.connect(self._on_sort_indicator_changed)
        self._list_view.itemClicked.connect(self._on_item_clicked)
        self._list_view.itemDoubleClicked.connect(self._on_item_double_clicked)
        self._tree_widget.itemClicked.connect(self._on_item_clicked)
        self._tree_widget.itemDoubleClicked.connect(self._on_item_double_clicked)

    def eventFilter(self, obj, event):
        if event.type() == QEvent.KeyPress:
            self.keyPressed.emit(event)

        return super(DataViewer, self).eventFilter(obj, event)

    def wheelEvent(self, event):
        """
        Overrides base wheelEvent function
        Triggered on any wheel events for the current widget
        :param event: QWheelEvent
        """

        modifier = QApplication.keyboardModifiers()
        valid_modifiers = (Qt.AltModifier, Qt.ControlModifier)
        if modifier in valid_modifiers:
            num_degrees = event.delta() / 8
            num_steps = num_degrees / 15
            delta = (num_steps * self.wheel_scroll_step())
            value = self.zoom_amount() + delta
            self.set_zoom_amount(value)

    def contextMenuEvent(self, event):
        """
        Overrides base contextMenuEvent function
        Shows context menu for this widget
        :param event: QEvent
        """

        menu = self.create_context_menu()
        point = QCursor.pos()
        return menu.exec_(point)

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def library(self):
        """
        Returns the library attached to the viewer
        :return: variant
        """

        return self._library

    def set_library(self, library):
        """
         Sets the data that will be showed in viewer
         :param library:
         """

        if library == self._library:
            return

        if not library and self._library:
            qtutils.safe_disconnect_signal(self._library.searchFinished)

        self._library = library

        if self._library:
            self.set_column_labels(library.field_names())
            library.searchFinished.connect(self._on_update_items)

    def library_window(self):
        """
        Returns library window this viewer belongs to
        :return: LibraryWindow
        """

        return self._library_window

    def label_display_option(self):
        """
        Returns the label display option
        :return: LabelDisplayOption
        """

        return self._label_display_option

    def set_label_display_option(self, value):
        """
        Sets the label display options
        :param value: LabelDisplayOption
        """

        self._label_display_option = value
        self.refresh_size()

    def is_table_view(self):
        """
        Returns whether widget is in list mode or not
        :return: bool
        """

        return not self._tree_widget.isHidden()

    def is_icon_view(self):
        """
        Returns whether widget is in icon mode or not
        :return: bool
        """

        return not self._list_view.isHidden()

    def set_view_mode(self, mode):
        """
        Sets the view mode for this widget
        :param mode: str
        """

        if mode == self.IconMode:
            self.set_zoom_amount(self.DEFAULT_MIN_ICON_SIZE)
        elif mode == self.TableMode:
            self.set_zoom_amount(self.DEFAULT_MIN_ICON_SIZE)

    def set_list_mode(self):
        """
        Sets the tree widget visible
        """

        self._list_view.hide()
        self._tree_widget.show()
        self._tree_widget.setFocus()

    def set_icon_mode(self):
        """
        Sets the list view visible
        """

        self._tree_widget.hide()
        self._list_view.show()
        self._list_view.setFocus()

    def zoom_amount(self):
        """
        Returns the zoom amount for the widget
        :return: int
        """

        return self._zoom_amount

    def set_zoom_amount(self, value):
        """
        Sets the zoom amount for the widget
        :param value: int
        """

        if value < self.DEFAULT_MIN_LIST_SIZE:
            value = self.DEFAULT_MIN_LIST_SIZE

        self._zoom_amount = value
        dpi = float(self.dpi())
        size = QSize(int(value * dpi), int(value * dpi))
        self.set_icon_size(size)
        if value >= self.DEFAULT_MIN_LIST_SIZE:
            self._set_view_mode(self.IconMode)
        else:
            self._set_view_mode(self.TableMode)
        column_width = value * dpi + self.item_text_height()
        self._tree_widget.setIndentation(0)
        self._tree_widget.setColumnWidth(0, column_width)
        self.scroll_to_selected_item()
        self.show_toast_message('Size: {}%'.format(value))

    def vertical_scrollbar(self):
        """
        Returns the active vertical scroll bar
        :return: QScrollBar
        """

        if self.is_table_view():
            return self.tree_widget().verticalScrollBar()
        else:
            return self.list_view().verticalScrollBar()

    def visual_item_rect(self, item):
        """
        Returns the visual rect for the item
        :param item: LibraryItem
        :return: QRect
        """

        if self.is_table_view():
            visual_rect = self.tree_widget().visual_item_rect(item)
        else:
            index = self.tree_widget().index_from_item(item)
            visual_rect = self.list_view().visualRect(index)

        return visual_rect

    def is_item_visible(self, item):
        """
        Returns whether given item is visible or not
        :param item: LibraryItem
        :return: bool
        """

        height = self.height()
        item_rect = self.visual_item_rect(item)
        scroll_bar_y = self.vertical_scrollbar().value()
        y = (scroll_bar_y - item_rect.y()) + height

        return scroll_bar_y < y < scroll_bar_y + height

    def scroll_to_item(self, item):
        """
        Ensures the given item is visible
        :param item: LibraryItem
        """

        position = QAbstractItemView.PositionAtCenter
        if self.is_table_view():
            self.tree_widget().scroll_to_item(item, position)
        elif self.is_icon_view():
            self.list_view().scroll_to_item(item, position)

    def scroll_to_selected_item(self):
        """
        Ensures that selected item is visible
        """

        item = self.selected_item()
        if item:
            self.scroll_to_item(item)

    def item_at(self, pos):
        """
        Returns the current item at the given position
        :param pos: QPoint
        :return: LibraryItem
        """

        if self.is_icon_view():
            return self.list_view().item_at(pos)
        else:
            return self.tree_widget().item_at(pos)

    def item_data(self, column_labels):
        """
        Returns all column data for the given column labels
        :param column_labels: list(str)
        :return: dict
        """

        data = dict()
        for item in self.items():
            key = item.id()
            for column_label in column_labels:
                column = self.treeWidget().column_from_label(column_label)
                value = item.data(column, Qt.EditRole)
                data.setdefault(key, dict())
                data[key].setdefault(column_label, value)

        return data

    def set_item_data(self, data):
        """
        Sets the item data for all the curren items
        :param data: dict
        """

        for item in self.items():
            key = item.id()
            if key in data:
                item.set_item_data(data[key])

    def item_text_height(self):
        """
        Returns the height of the item text
        :return: int
        """

        return self.DEFAULT_TEXT_HEIGHT * self.dpi()

    def item_delegate(self):
        """
        Returns the item delegate for the views
        :return: LibraryViewerDelegate
        """

        return self._delegate

    def column_labels_from_items(self):
        """
        Returns the column labels for all the items
        :return: list(str)
        """

        seq = list()
        for item in self.items():
            seq.extend(item._text_column_order)
        seen = set()

        return [x for x in seq if x not in seen and not seen.add(x)]

    def padding(self):
        """
        Returns the item padding
        :return: int
        """

        return self._padding

    def set_padding(self, value):
        """
        Sets the item padding
        :param value: int
        """

        if value % 2 == 0:
            self._padding = value
        else:
            self._padding = value + 1
        self.repaint()
        self.show_toast_message('Border: {}'.format(value))

    def item_size_hint(self):
        """
        Returns the item size hint
        :return: QSize
        """

        return self._item_size_hint

    def icon_size(self):
        """
        Returns the icon size for the views
        :return: QSize
        """

        return self._icon_size

    def set_icon_size(self, size):
        """
        Sets the icon size for the views
        :param size: QSize
        """

        self._icon_size = size
        if self.label_display_option() == item.LabelDisplayOption.Under:
            w = size.width()
            h = size.width() + self.item_text_height()
            self._item_size_hint = QSize(w, h)
        elif size.height() < self.item_text_height():
            self._item_size_hint = QSize(size.width(), self.item_text_height())
        else:
            self._item_size_hint = size

        self.list_view().setIconSize(size)
        self.tree_widget().setIconSize(size)

    def wheel_scroll_step(self):
        """
        Returns the wheel scroll setp amount
        :return: int
        """

        return self.DEFAULT_WHEEL_SCROLL_STEP

    def refresh(self):
        """
        Refreshes the item size
        """

        self.refresh_size()

    def refresh_columns(self):
        """
        Refresh columns labels
        """

        self.set_column_labels(self.column_labels_from_items())

    def refresh_size(self):
        """
        Refreshes the size of the items
        """

        self.set_zoom_amount(self.zoom_amount() + 1)
        self.set_zoom_amount(self.zoom_amount() - 1)
        self.repaint()

    # ============================================================================================================
    # DPI
    # ============================================================================================================

    def dpi(self):
        """
        Returns zoom multiplier
        :return: int
        """

        return self._dpi

    def set_dpi(self, dpi):
        """
        Sets the zoom multiplier
        :param dpi: int
        """

        self._dpi = dpi
        self.refresh_size()

    # ============================================================================================================
    # TREE VIEW
    # ============================================================================================================

    def tree_widget(self):
        """
        Returns the list view that contains the items
        :return: TreeWidget
        """

        return self._tree_widget

    def column_from_label(self, *args):
        """
        Returns column from given label text
        :return: int
        """

        return self.tree_widget().column_from_label(*args)

    def update_columns(self):
        """
        Updates the columns labels with the current item data
        """

        self.treeWidget().update_header_labels()

    def set_column_hidden(self, column, hidden):
        """
        Hides/Shows specific column in tree widget
        :param column: int
        :param hidden: bool
        """

        self.tree_widget().setColumnHidden(column, hidden)

    def column_labels(self):
        """
        Returns all the column labels
        :return: list(str)
        """

        return self.treeWidget().column_labels()

    def set_column_labels(self, labels):
        """
        Set the columns for the viewer
        :param labels: list(str)
        """

        labels_set = set()
        set_add = labels_set.add
        labels = [x for x in labels if x.strip() and not (x in labels_set or set_add(x))]

        if 'Custom Order' not in labels:
            labels.append('Custom Order')

        self.tree_widget().setHeaderLabels(labels)

        self.set_column_hidden(labels.index('Custom Order'), True)

    def index_from_item(self, item):
        """
        Returns the QModelIndex associated with the given item
        :param item: LibraryItem
        :return: QModelIndex
        """

        return self._tree_widget.index_from_item(item)

    def items(self):
        """
        Returns all the items in the tree widget
        :return: list(Item)
        """

        return self._tree_widget.items()

    def item_from_index(self, index):
        """
        Returns a pointer to the QTreeWidgetItem associated with the given index
        :param index: QModelIndex
        :return: QTreeWidgetItem
        """

        return self.tree_widget().itemFromIndex(index)

    def text_from_items(self, *args, **kwargs):
        """
        Returns all data for the given items and given column
        :return: list(str)
        """

        return self.tree_widget().text_from_items(*args, **kwargs)

    def text_from_column(self, *args, **kwargs):
        """
        Returns all data for the given column
        :return: list(str)
        """

        return self.tree_widget().text_from_column(*args, **kwargs)

    def add_item(self, item):
        """
        Add the item to the tree widget
        :param item: LibraryItem
        """

        self.add_items([item])

    def add_items(self, items):
        """
        Add the given items to the items widget
        :param items: list(LibraryItem)
        """

        self._tree_widget.addTopLevelItems(items)

    def insert_items(self, items, item_at=None):
        """
        Inserts the given items at the given position
        :param items: list(LibraryItem)
        :param item_at: LibraryItem
        """

        self.add_items(items)
        self.move_items(items, item_at=item_at)
        self.tree_widget().set_items_selected(items)

    def selected_item(self):
        """
        Returns the last non-hidden selected item
        :return: LibraryItem
        """

        return self._tree_widget.selected_item()

    def selected_items(self):
        """
        Returns a list with all selected non-hiden items
        :return: list(QTreeWidgetItem)
        """

        return self._tree_widget.selectedItems()

    def set_item_hidden(self, item, value):
        """
        Sets the visibilty of given item
        :param item: QTreeWidgetItem
        :param value: bool
        """

        item.setHidden(value)

    def set_items_hidden(self, items, value):
        """
        Set the visibility of given items
        :param items: list(QTreeWidgetItem)
        :param value: bool
        """

        for item in items:
            self.set_item_hidden(item, value)

    def selected_paths(self):
        """
        Returns the selected item paths
        :return: list(str)
        """

        paths = list()
        for item in self.selected_items():
            path = item.url().toLocalFile()
            paths.append(path)

        return paths

    def select_paths(self, paths):
        """
        Selected the items that have the given paths
        :param paths: list(str)
        """

        for item in self.items():
            path = item.path()
            if path in paths:
                item.setSelected(True)

    def select_items(self, items):
        """
        Select the given items
        :param items: list(LibraryItem)
        """

        paths = [item.path() for item in items]
        self.select_paths(paths)

    def clear_selection(self):
        """
        Cleras the user selection
        """

        self._tree_widget.clearSelection()

    def selection_model(self):
        """
        Returns the current selection model
        :return: QItemSelectionModel
        """

        return self._tree_widget.selectionModel()

    def model(self):
        """
        Returns the model the viewer is representing
        :return: QAbstractItemModel
        """

        return self._tree_widget.model()

    def create_group_item(self, text, children=None):
        """
        Internal function that creates a new item for the given text and children
        :param text: str
        :param children: list(LibraryItem)
        """

        group_item = group.GroupDataItemView()
        group_item.set_name(text)
        group_item.set_stretch_to_widget(self)
        group_item.set_children(children)

        return group_item

    def update_items(self):
        """
        Sets the items to the viewer
        """

        selected_items = self.selected_items()

        with qt_contexts.block_signals(self.tree_widget()):

            try:
                self.clear_selection()
                if self.library():
                    results = self.library().grouped_results()
                    item_views = list()
                    for group in results:
                        if group != 'None':
                            group_item = self.create_group_item(group)
                            item_views.append(group_item)
                        for item in results[group]:
                            view_class = self.library_window().factory.get_view(item)
                            if not view_class:
                                continue
                            item_view = view_class(item, library_window=self.library_window())
                            item_views.append(item_view)

                    if item_views:
                        self.tree_widget().set_items(item_views)
                        if selected_items:
                            self.select_items(selected_items)
                            self.scroll_to_selected_item()
                    else:
                        self.clear()
                else:
                    self.clear()
            finally:
                self.itemSelectionChanged.emit()

    def clear(self):
        """
        Clear all elements in tree widget
        """

        self.tree_widget().clear()

    # ============================================================================================================
    # LIST VIEW
    # ============================================================================================================

    def list_view(self):
        """
        Returns the list view that contains the items
        :return: LibraryListView
        """

        return self._list_view

    def set_locked(self, value):
        """
        Disables drag and drop
        :param value: bool
        """

        self.list_view().setDragEnabled(not value)
        self.list_view().setDropEnabled(not value)

    def spacing(self):
        """
        Returns the spacing between the items
        :return: int
        """

        return self.list_view().spacing()

    def set_spacing(self, spacing):
        """
        Set the spacing between the items
        :param spacing: int
        """

        self.list_view().setSpacing(spacing)
        self.scroll_to_selected_item()
        self.show_toast_message('Spacing: {}'.format(spacing))

    def move_items(self, items, item_at=None):
        """
        Moves the given items to the given position
        :param items: list(LibraryItem)
        :param item_at: LibraryItem
        """

        self.list_view().move_items(items, item_at=item_at)

    # ============================================================================================================
    # TOAST WIDGET
    # ============================================================================================================

    def toast_enabled(self):
        """
        Returns whether toast message widget is enabled or not
        :return: bool
        """

        return self._toast_enabled

    def set_toast_enabled(self, flag):
        """
        Sets whether toast widget is enabled or not
        :param flag: bool
        """

        self._toast_enabled = flag

    def show_toast_message(self, text, duration=300):
        """
        Shows a toast with the given text for the given duration
        :param text: str
        :param duration: None or int
        """

        if self.toast_enabled():
            self._toast_widget.set_duration(duration)
            self._toast_widget.setText(text)
            self._toast_widget.show()

    # ============================================================================================================
    # MENUS
    # ============================================================================================================

    def create_settings_menu(self):
        """
        Creates and returns the settings menu for this widget
        :return: QMenu
        """

        menu = QMenu('Item View', self)
        menu.setIcon(resources.icon('eye'))
        menu.addSeparator()

        # copy_text_menu = self.tree_widget().create_copy_text_menu()
        # menu.addMenu(copy_text_menu)
        # menu.addSeparator()

        size_action = action.SliderAction('Size', menu)
        size_action.slider().setMinimum(10)
        size_action.slider().setMaximum(200)
        size_action.slider().setValue(self.zoom_amount())
        size_action.slider().valueChanged.connect(self.set_zoom_amount)
        menu.addAction(size_action)

        border_action = action.SliderAction('Border', menu)
        border_action.slider().setMinimum(0)
        border_action.slider().setMaximum(20)
        border_action.slider().setValue(self.padding())
        border_action.slider().valueChanged.connect(self.set_padding)
        menu.addAction(border_action)

        spacing_action = action.SliderAction('Spacing', menu)
        spacing_action.slider().setMinimum(self.DEFAULT_MIN_SPACING)
        spacing_action.slider().setMaximum(self.DEFAULT_MAX_SPACING)
        spacing_action.slider().setValue(self.spacing())
        spacing_action.slider().valueChanged.connect(self.set_spacing)
        menu.addAction(spacing_action)

        menu.addAction(action.SeparatorAction('Label Options', menu))

        for option in item.LabelDisplayOption.values():
            option_action = QAction(option.title(), menu)
            option_action.setCheckable(True)
            option_action.setChecked(option == self.label_display_option())
            option_callback = partial(self.set_label_display_option, option)
            option_action.triggered.connect(option_callback)
            menu.addAction(option_action)

        return menu

    def create_items_menu(self, items=None):
        """
        Creates the item menu for the given items
        :param items: list(LibraryItem)
        :return: QMenu
        """

        item = items or self.selected_item()
        menu = QMenu(self)
        if item:
            try:
                item.context_menu(menu)
            except Exception as e:
                LOGGER.exception(e)
        else:
            item_action = QAction(menu)
            item_action.setText('No Item selected')
            item_action.setDisabled(True)
            menu.addAction(item_action)

        return menu

    def create_copy_text_menu(self):
        """
        Returns copy text menu for this widget
        :return: QMenu
        """

        return self.tree_widget().create_copy_text_menu()

    def create_context_menu(self):
        """
        Creates and returns the context menu for this widget
        :return: QMenu
        """

        menu = self.create_items_menu()
        settings_menu = self.create_settings_menu()
        menu.addMenu(settings_menu)

        return menu

    # ============================================================================================================
    # PAINT
    # ============================================================================================================

    def text_color(self):
        """
        Returns the item text color
        :return: QColor
        """

        return self._text_color

    def set_text_color(self, color):
        """
        Sets the item text color
        :param color: QColor
        """

        self._text_color = color

    def text_selected_color(self):
        """
        Returns the item text color when selected
        :return: QColor
        """

        return self._text_selected_color

    def set_text_selected_color(self, color):
        """
        Sets the text color when an item is selected
        :param color: QColor
        """

        self._text_selected_color = color

    def background_color(self):
        """
        Returns item background color
        :return: QColor
        """

        return self._background_color

    def set_background_color(self, color):
        """
        Sets the item background color
        :param color: QColor
        """

        self._background_color = color

    def background_hover_color(self):
        """
        Returns the background color when the mouse is over an item
        :return: QColor
        """

        return self._background_hover_color

    def set_background_hover_color(self, color):
        """
        Sets the background color when the mouse hovers over the item
        :param color: QColor
        """

        self._background_hover_color = color

    def background_selected_color(self):
        """
        Returns the background color when an item is selected
        :return: QColor
        """

        return self._background_selected_color

    def set_background_selected_color(self, color):
        """
        Sets the background color when an item is selected
        :param color: QColor
        """

        self._background_selected_color = color
        self._list_view.ser_rubber_band_color(QColor(200, 200, 200, 255))

    # ============================================================================================================
    # SETTINGS
    # ============================================================================================================

    def settings(self):
        """
        Returns the current state of the widget
        :return: dict
        """

        settings = dict()
        settings['columnLabels'] = self.column_labels()
        settings['padding'] = self.padding()
        settings['spacing'] = self.spacing()
        settings['zoomAmount'] = self.zoom_amount()
        settings['selectedPaths'] = self.selected_paths()
        settings['labelDisplayOption'] = self.label_display_option()
        settings.update(self.tree_widget().settings())

        return settings

    def set_settings(self, settings):
        """
        Sets the current state of the widget
        :param settings: dict
        """

        self.set_toast_enabled(False)
        padding = settings.get('padding', 5)
        spacing = settings.get('spacing', 2)
        zoom_amount = settings.get('zoomAmount', 100)
        selected_paths = settings.get('selectedPaths', list())
        label_display_option = settings.get('labelDisplayOption', self.LABEL_DISPLAY_OPTION)
        self.set_padding(padding)
        self.set_spacing(spacing)
        self.set_zoom_amount(zoom_amount)
        self.select_paths(selected_paths)
        self.set_label_display_option(label_display_option)
        self.tree_widget().set_settings(settings)
        self.set_toast_enabled(True)

        return settings

    # ============================================================================================================
    # INTERNAL
    # ============================================================================================================

    def _set_view_mode(self, mode):
        """
        Internal function that sets the view mode ro this widget
        :param mode: str
        :return: str
        """

        if mode == self.IconMode:
            self.set_icon_mode()
        elif mode == self.TableMode:
            self.set_list_mode()

    # ============================================================================================================
    # CALLBACKS
    # ============================================================================================================

    def _on_sort_indicator_changed(self):
        """
        Internal callback function that is called when sort indicator changes
        """

        pass

    def _on_item_clicked(self, item):
        """
        Internal callback function that is called when an item has been clicked
        :param item: DataItem
        """

        if isinstance(item, group.GroupDataItemView):
            self.groupClicked.emit(item)
        else:
            self.itemClicked.emit(item)

    def _on_item_double_clicked(self, item):
        """
        Internal callback function that is called when an item has been double clicked
        :param item: DataItem
        """

        self.itemDoubleClicked.emit(item)

    def _on_update_items(self):
        """
        Internal callback function that sets the items to the viewer
        """

        self.update_items()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains library item tree view implementation
"""

from __future__ import print_function, division, absolute_import

import logging
from functools import partial

from Qt.QtCore import Qt
from Qt.QtWidgets import QApplication, QTreeWidget, QAbstractItemView, QMenu
from Qt.QtGui import QCursor, QFontMetrics, QClipboard

from tpDcc.libs.python import python

from tpDcc.tools.datalibrary.core import consts
from tpDcc.tools.datalibrary.data import group
from tpDcc.tools.datalibrary.widgets import mixinview

LOGGER = logging.getLogger('tpDcc-libs-datalibrary')


class ViewerTreeView(mixinview.ViewerViewWidgetMixin, QTreeWidget):
    def __init__(self, *args, **kwargs):
        QTreeWidget.__init__(self, *args, **kwargs)
        mixinview.ViewerViewWidgetMixin.__init__(self)

        self._header_labels = list()
        self._hidden_columns = dict()

        self.setAutoScroll(False)
        self.setMouseTracking(True)
        self.setSortingEnabled(False)
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)

        header = self.header()
        header.setContextMenuPolicy(Qt.CustomContextMenu)
        header.customContextMenuRequested.connect(self._on_show_header_menu)

        self.itemClicked.connect(self._on_item_clicked)
        self.itemDoubleClicked.connect(self._on_item_double_clicked)

    # ============================================================================================================
    # OVERRIDES - MIXIN
    # ============================================================================================================

    def mouseMoveEvent(self, event):
        """
        Triggered when the user moves the mouse over the current viewport
        :param event: QMouseEvent
        """

        mixinview.ViewerViewWidgetMixin.mouseMoveEvent(self, event)
        QTreeWidget.mouseMoveEvent(self, event)

    def mouseReleaseEvent(self, event):
        """
        Triggered when the user releases the mouse button on the viewport
        :param event: QMouseEvent
        """

        mixinview.ViewerViewWidgetMixin.mouseReleaseEvent(self, event)
        QTreeWidget.mouseReleaseEvent(self, event)

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def drawRow(self, painter, options, index):
        """
        Overrides base QTreeWidget drawDrow function
        :param painter: QPainter
        :param options: QStyleOption
        :param index: QModelIndex
        """

        item = self.itemFromIndex(index)
        item.paint_row(painter, options, index)

    def items(self):
        """
        Overrides base QTreeWidget items function
        Return a list of all items in the tree widget
        :return: list(LibraryItem)
        """

        items_list = list()
        for item in self._items():
            if not isinstance(item, group.GroupDataItemView):
                items_list.append(item)

        return items_list

    def selectedItems(self):
        """
        Overrides base QTreeWidget selectedItems function
        Returns all selected items
        :return: list(LibraryItem)
        """

        items_list = list()
        items_ = super(ViewerTreeView, self).selectedItems()

        for item in items_:
            if not isinstance(item, group.GroupDataItemView):
                items_list.append(item)

        return items_list

    def setColumnHidden(self, column, value):
        """
        Overrides base QTreeWidget setColumnHidden function
        :param column: int or str
        :param value: bool
        """

        if python.is_string(column):
            column = self.column_from_label(column)

        label = self.label_from_column(column)
        self._hidden_columns[label] = value

        super(ViewerTreeView, self).setColumnHidden(column, value)

        width = self.columnWidth(column)
        if width < consts.TREE_MINIMUM_WIDTH:
            width = consts.TREE_DEFAULT_WIDTH
            self.setColumnWidth(column, width)

    def resizeColumnToContents(self, column):
        """
        Overrides base QTreeWidget resizeColumnToContents function
        Resize the given column to the data of that column
        :param column: int or str
        """

        width = 0
        for item in self.items():
            text = item.text(column)
            font = item.font(column)
            metrics = QFontMetrics(font)
            text_width = metrics.width(text) + item.padding()
            width = max(width, text_width)

        self.setColumnWidth(column, width)

    def setHeaderLabels(self, labels):
        """
        Overrides base QTreeWidget setHeaderLabels function
        :param labels: list(str)
        """

        s = set()
        sadd = s.add
        labels = [x for x in labels if x.strip() and not (x in s or sadd(x))]

        if 'Custom Order' not in labels:
            labels.append('Custom Order')
        self.setColumnHidden('Custom Order', True)

        column_settings = self.column_settings()
        super(ViewerTreeView, self).setHeaderLabels(labels)
        self._header_labels = labels
        self.update_column_hidden()
        self.set_column_settings(column_settings)

    def clear(self, *args):
        """
        Overrides base QTreeWidget clear function
        Clear all dirty tree items
        """

        super(ViewerTreeView, self).clear(*args)
        self.clean_dirty_objects()

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def set_items(self, items):
        """
        Add given items to the tree, clearing the tree first
        :param items: list(LibraryItem)
        """

        selected_items = self.selectedItems()
        self.take_top_level_items()
        self.addTopLevelItems(items)
        self.set_items_selected(selected_items, True)

    def set_items_selected(self, items, value, scroll_to=True):
        """
        Selects the given library items
        :param items: list(LibraryItem)
        :param value: bool, Whether to select or deselect the items
        :param scroll_to: bool, Whether to scroll or not to selected items
        """

        for item in items:
            self.setItemSelected(item, value)
        if scroll_to:
            self.viewer().scroll_to_selected_item()

    def selected_item(self):
        """
        Returns the last non-hidden selected item
        :return: LibraryItem
        """

        items = self.selectedItems()
        if items:
            return items[-1]

        return None

    def take_top_level_items(self):
        """
        Returns all items from the tree widget
        :return: list(LibraryItem)
        """

        items_list = list()
        for item in self._items():
            # It is faster to take from first index
            items_list.append(self.takeTopLevelItem(1))
        items_list.append(self.takeTopLevelItem(0))

        return items_list

    def text_from_items(self, items, column, split=None, duplicates=False):
        """
        Returns all the text data for the given items and column
        :param items: list(LibraryItem)
        :param column: int or str
        :param split: str
        :param duplicates: bool
        :return: list(str)
        """

        results = list()

        for item in items:
            text = item.text(column)
            if text and split:
                results.extend(text.split(split))
            elif text:
                results.append(text)
        if not duplicates:
            results = list(set(results))

        return results

    def text_from_column(self, column, split=None, duplicates=False):
        """
        Returns all data for the given column
        :param column: int or str
        :param split: str
        :param duplicates: bool
        :return: list(str)
        """

        items_list = self.items()
        results = self.text_from_items(items_list, column, split=split, duplicates=duplicates)

        return results

    def header_labels(self):
        """
        Returns all header labels
        :return: list(str)
        """

        return self._header_labels

    def is_header_label(self, label):
        """
        Returns whether given label is a valid header label or not
        :param label: str
        :return: bool
        """

        return label in self._header_labels

    def column_labels(self):
        """
        Returns all header labels for the tree widget
        :return: list(str)
        """

        return self.header_labels()

    def column_from_label(self, label):
        """
        Returns the column for the given label
        :param label: str
        :return: int
        """

        try:
            return self._header_labels.index(label)
        except ValueError:
            return -1

    def label_from_column(self, column):
        """
        Returns the column label for the given column
        :param column: int
        :return: str
        """

        if column is not None:
            return self.headerItem().text(column)

    def item_row(self, item):
        """
        Returns the row for the given item
        :param item: LibraryItem
        :return: int
        """

        index = self.indexFromItem(item)
        return index.row()

    def row_at(self, pos):
        """
        Returns the row for the given position
        :param pos: QPoint
        :return: int
        """

        item = self.itemAt(pos)
        return self.item_row(item)

    def show_all_columns(self):
        """
        Show all available columns
        """

        for column in range(self.columnCount()):
            self.setColumnHidden(column, False)

    def hide_all_columns(self):
        """
        Hide all available columns
        """

        for column in range(1, self.columnCount()):
            self.setColumnHidden(column, True)

    def update_column_hidden(self):
        """
        Updates the hidden state for all the current columns
        """

        self.show_all_columns()
        column_labels = list(self._hidden_columns.keys())
        for column_label in column_labels:
            self.setColumnHidden(column_label, self._hidden_columns[column_label])

    # ============================================================================================================
    # CUSTOM ORDER
    # ============================================================================================================

    def is_sort_by_custom_order(self):
        """
        Returns True if items are currently sorted by custom order
        :return: bool
        """

        return 'Custom Order' in str(self.viewer().library().sort_by())

    def items_custom_order(self):
        """
        Returns the items sorted by hte custom order data
        :return: list(DataItem)
        """

        items = self.items()

        def sort_key(item):
            return item.item_data().get('Custom Order', 0)

        custom_order = sorted(items, key=sort_key)

        return custom_order

    def set_items_custom_order(self, items, row=1, padding=5):
        """
        Sets the custom order for the given items by their list order
        :param items:
        :param row:
        :param padding:
        :return:
        """

        for item in items:
            item.item_data()['Custom Order'] = str(row).zfill(padding)
            row += 1

    def move_items(self, items, item_at):
        """
        Moves the given items to the position of the target row
        :param items: list(DataItem)
        :param item_at: DataItem
        """

        row = 0
        ordered_items = self.items_custom_order()
        if item_at:
            row = ordered_items.index(item_at)
        removed_items = list()

        for item in items:
            index = ordered_items.index(item)
            removed_items.append(ordered_items.pop(index))

        for item in removed_items:
            ordered_items.insert(row, item)

        self.set_items_custom_order(ordered_items)

        for item in items:
            item.setSelected(True)

        self.viewer().scroll_to_selected_item()

    # ============================================================================================================
    # CLIPBOARD
    # ============================================================================================================

    def copy_text(self, column):
        """
        Copy the given column text to clipboard
        :param column: int or text
        """

        items = self.selectedItems()
        text = '\n'.join([item.text(column) for item in items])
        clipboard = QApplication.clipboard()
        clipboard.setText(text, QClipboard.Clipboard)

    def create_copy_text_menu(self):
        """
        Creates a menu to cpoy the selected item data to the clipboard
        :return: QMenu
        """

        menu = QMenu('Copy Text', self)
        if self.selectedItems():
            for column in range(self.columnCount()):
                label = self.label_from_column(column)
                action = menu.addAction(label)
                action_callback = partial(self.copy_text, column)
                action.triggered.connect(action_callback)
        else:
            action = menu.addAction('No items selected')
            action.setEnabled(False)

        return menu

    # ============================================================================================================
    # MENUS
    # ============================================================================================================

    def show_header_menu(self, pos):
        """
        Creates and show the header menu at the cursor position
        :param pos: QPoint
        :return: QMenu
        """

        header = self.header()
        column = header.logicalIndexAt(pos)
        menu = self._create_header_menu(column)
        menu.addSeparator()
        sub_menu = self._create_hide_column_menu()
        menu.addMenu(sub_menu)
        menu.exec_(QCursor.pos())

    # ============================================================================================================
    # SETTINGS
    # ============================================================================================================

    def settings(self):
        """
        Returns the current widget settings
        :return: dict
        """

        settings = dict()
        settings['columnSettings'] = self.column_settings()

        return settings

    def set_settings(self, settings):
        """
        Sets the current widget settings
        :param settings: dict
        :return: dict
        """

        column_settings = settings.get('columnSettings', dict())
        self.set_column_settings(column_settings)

        return settings

    def column_settings(self):
        """
        Returns the settings for each column
        :return: dict
        """

        column_settings = dict()

        for column in range(self.columnCount()):
            label = self.label_from_column(column)
            hidden = self.isColumnHidden(column)
            width = self.columnWidth(column)
            column_settings[label] = {
                'index': column,
                'hidden': hidden,
                'width': width
            }

        return column_settings

    def set_column_settings(self, settings):
        """
        Set the settings for each column
        :param settings: dict
        """

        for label in settings:
            if self.is_header_label(label):
                column = self.column_from_label(label)
                width = settings[label].get('width', 100)
                if width < consts.TREE_MINIMUM_WIDTH:
                    width = consts.TREE_DEFAULT_WIDTH
                self.setColumnWidth(column, width)
                hidden = settings[label].get('hidden', False)
                self.setColumnHidden(column, hidden)
            else:
                LOGGER.debug('Cannot set the column setting for header label: {}'.format(label))

    # ============================================================================================================
    # INTERNAL
    # ============================================================================================================

    def _items(self):
        """
        Internal function that returns a list of all items in the tree widget
        :return: list(LibraryItem)
        """

        return self.findItems('*', Qt.MatchWildcard | Qt.MatchRecursive)

    def _create_header_menu(self, column):
        """
        Internal function that creates a new header menu
        :param column, iht
        :return: QMenu
        """

        menu = QMenu(self)
        label = self.label_from_column(column)
        hide_action = menu.addAction('Hide "{}"'.format(label))
        hide_callback = partial(self.setColumnHidden, column, True)
        hide_action.triggered.connect(hide_callback)
        menu.addSeparator()
        resize_action = menu.addAction('Resize to Contents')
        resize_callback = partial(self.resizeColumnToContents, column)
        resize_action.triggered.connect(resize_callback)

        return menu

    def _create_hide_column_menu(self):
        """
        Internal function that creates the hide column menu
        :return: QMenu
        """

        menu = QMenu('Show/Hide Column', self)
        show_all_action = menu.addAction('Show All')
        show_all_action.triggered.connect(self.show_all_columns)
        hide_all_action = menu.addAction('Hide All')
        hide_all_action.triggered.connect(self.hide_all_columns)
        menu.addSeparator()
        for column in range(self.columnCount()):
            label = self.label_from_column(column)
            is_hidden = self.isColumnHidden(column)
            action = menu.addAction(label)
            action.setCheckable(True)
            action.setChecked(not is_hidden)
            action_callback = partial(self.setColumnHidden, column, not is_hidden)
            action.triggered.connect(action_callback)

        return menu

    # ============================================================================================================
    # CALLBACKS
    # ============================================================================================================

    def _on_show_header_menu(self):
        pass

    def _on_item_clicked(self, item):
        item.clicked()

    def _on_item_double_clicked(self, item):
        item.double_clicked()
